from __future__ import unicode_literals
from moto.core.responses import BaseResponse


class ReservedInstances(BaseResponse):

    def cancel_reserved_instances_listing(self):
        if self.is_not_dryrun('CancelReservedInstances'):
            raise NotImplementedError(
                'ReservedInstances.cancel_reserved_instances_listing is not yet implemented')

    def create_reserved_instances_listing(self):
        if self.is_not_dryrun('CreateReservedInstances'):
            raise NotImplementedError(
                'ReservedInstances.create_reserved_instances_listing is not yet implemented')

    def describe_reserved_instances(self):
        raise NotImplementedError(
            'ReservedInstances.describe_reserved_instances is not yet implemented')

    def describe_reserved_instances_listings(self):
        raise NotImplementedError(
            'ReservedInstances.describe_reserved_instances_listings is not yet implemented')

    def describe_reserved_instances_offerings(self):
        raise NotImplementedError(
            'ReservedInstances.describe_reserved_instances_offerings is not yet implemented')

    def purchase_reserved_instances_offering(self):
        if self.is_not_dryrun('PurchaseReservedInstances'):
            raise NotImplementedError(
                'ReservedInstances.purchase_reserved_instances_offering is not yet implemented')



from __future__ import unicode_literals

from collections import defaultdict

from moto.core import BaseBackend, BaseModel
from moto.core.exceptions import RESTError
from moto.ec2 import ec2_backends
from moto.cloudformation import cloudformation_backends

import datetime
import time
import uuid
import itertools


class Parameter(BaseModel):
    def __init__(self, name, value, type, description, allowed_pattern, keyid,
            last_modified_date, version):
        self.name = name
        self.type = type
        self.description = description
        self.allowed_pattern = allowed_pattern
        self.keyid = keyid
        self.last_modified_date = last_modified_date
        self.version = version

        if self.type == 'SecureString':
            self.value = self.encrypt(value)
        else:
            self.value = value

    def encrypt(self, value):
        return 'kms:{}:'.format(self.keyid or 'default') + value

    def decrypt(self, value):
        if self.type != 'SecureString':
            return value

        prefix = 'kms:{}:'.format(self.keyid or 'default')
        if value.startswith(prefix):
            return value[len(prefix):]

    def response_object(self, decrypt=False):
        r = {
            'Name': self.name,
            'Type': self.type,
            'Value': self.decrypt(self.value) if decrypt else self.value,
            'Version': self.version,
        }

        return r

    def describe_response_object(self, decrypt=False):
        r = self.response_object(decrypt)
        r['LastModifiedDate'] = int(self.last_modified_date)
        r['LastModifiedUser'] = 'N/A'

        if self.description:
            r['Description'] = self.description

        if self.keyid:
            r['KeyId'] = self.keyid

        if self.allowed_pattern:
            r['AllowedPattern'] = self.allowed_pattern

        return r


MAX_TIMEOUT_SECONDS = 3600


class Command(BaseModel):
    def __init__(self, comment='', document_name='', timeout_seconds=MAX_TIMEOUT_SECONDS,
            instance_ids=None, max_concurrency='', max_errors='',
            notification_config=None, output_s3_bucket_name='',
            output_s3_key_prefix='', output_s3_region='', parameters=None,
            service_role_arn='', targets=None, backend_region='us-east-1'):

        if instance_ids is None:
            instance_ids = []

        if notification_config is None:
            notification_config = {}

        if parameters is None:
            parameters = {}

        if targets is None:
            targets = []

        self.error_count = 0
        self.completed_count = len(instance_ids)
        self.target_count = len(instance_ids)
        self.command_id = str(uuid.uuid4())
        self.status = 'Success'
        self.status_details = 'Details placeholder'

        self.requested_date_time = datetime.datetime.now()
        self.requested_date_time_iso = self.requested_date_time.isoformat()
        expires_after = self.requested_date_time + datetime.timedelta(0, timeout_seconds)
        self.expires_after = expires_after.isoformat()

        self.comment = comment
        self.document_name = document_name
        self.instance_ids = instance_ids
        self.max_concurrency = max_concurrency
        self.max_errors = max_errors
        self.notification_config = notification_config
        self.output_s3_bucket_name = output_s3_bucket_name
        self.output_s3_key_prefix = output_s3_key_prefix
        self.output_s3_region = output_s3_region
        self.parameters = parameters
        self.service_role_arn = service_role_arn
        self.targets = targets
        self.backend_region = backend_region

        # Get instance ids from a cloud formation stack target.
        stack_instance_ids = [self.get_instance_ids_by_stack_ids(target['Values']) for
            target in self.targets if
            target['Key'] == 'tag:aws:cloudformation:stack-name']

        self.instance_ids += list(itertools.chain.from_iterable(stack_instance_ids))

        # Create invocations with a single run command plugin.
        self.invocations = []
        for instance_id in self.instance_ids:
            self.invocations.append(
                self.invocation_response(instance_id, "aws:runShellScript"))

    def get_instance_ids_by_stack_ids(self, stack_ids):
        instance_ids = []
        cloudformation_backend = cloudformation_backends[self.backend_region]
        for stack_id in stack_ids:
            stack_resources = cloudformation_backend.list_stack_resources(stack_id)
            instance_resources = [
                instance.id for instance in stack_resources
                if instance.type == "AWS::EC2::Instance"]
            instance_ids.extend(instance_resources)

        return instance_ids

    def response_object(self):
        r = {
            'CommandId': self.command_id,
            'Comment': self.comment,
            'CompletedCount': self.completed_count,
            'DocumentName': self.document_name,
            'ErrorCount': self.error_count,
            'ExpiresAfter': self.expires_after,
            'InstanceIds': self.instance_ids,
            'MaxConcurrency': self.max_concurrency,
            'MaxErrors': self.max_errors,
            'NotificationConfig': self.notification_config,
            'OutputS3Region': self.output_s3_region,
            'OutputS3BucketName': self.output_s3_bucket_name,
            'OutputS3KeyPrefix': self.output_s3_key_prefix,
            'Parameters': self.parameters,
            'RequestedDateTime': self.requested_date_time_iso,
            'ServiceRole': self.service_role_arn,
            'Status': self.status,
            'StatusDetails': self.status_details,
            'TargetCount': self.target_count,
            'Targets': self.targets,
        }

        return r

    def invocation_response(self, instance_id, plugin_name):
        # Calculate elapsed time from requested time and now. Use a hardcoded
        # elapsed time since there is no easy way to convert a timedelta to
        # an ISO 8601 duration string.
        elapsed_time_iso = "PT5M"
        elapsed_time_delta = datetime.timedelta(minutes=5)
        end_time = self.requested_date_time + elapsed_time_delta

        r = {
            'CommandId': self.command_id,
            'InstanceId': instance_id,
            'Comment': self.comment,
            'DocumentName': self.document_name,
            'PluginName': plugin_name,
            'ResponseCode': 0,
            'ExecutionStartDateTime': self.requested_date_time_iso,
            'ExecutionElapsedTime': elapsed_time_iso,
            'ExecutionEndDateTime': end_time.isoformat(),
            'Status': 'Success',
            'StatusDetails': 'Success',
            'StandardOutputContent': '',
            'StandardOutputUrl': '',
            'StandardErrorContent': '',
        }

        return r

    def get_invocation(self, instance_id, plugin_name):
        invocation = next(
            (invocation for invocation in self.invocations
                if invocation['InstanceId'] == instance_id), None)

        if invocation is None:
            raise RESTError(
                'InvocationDoesNotExist',
                'An error occurred (InvocationDoesNotExist) when calling the GetCommandInvocation operation')

        if plugin_name is not None and invocation['PluginName'] != plugin_name:
                raise RESTError(
                    'InvocationDoesNotExist',
                    'An error occurred (InvocationDoesNotExist) when calling the GetCommandInvocation operation')

        return invocation


class SimpleSystemManagerBackend(BaseBackend):

    def __init__(self):
        self._parameters = {}
        self._resource_tags = defaultdict(lambda: defaultdict(dict))
        self._commands = []

        # figure out what region we're in
        for region, backend in ssm_backends.items():
            if backend == self:
                self._region = region

    def delete_parameter(self, name):
        try:
            del self._parameters[name]
        except KeyError:
            pass

    def delete_parameters(self, names):
        result = []
        for name in names:
            try:
                del self._parameters[name]
                result.append(name)
            except KeyError:
                pass
        return result

    def get_all_parameters(self):
        result = []
        for k, _ in self._parameters.items():
            result.append(self._parameters[k])
        return result

    def get_parameters(self, names, with_decryption):
        result = []
        for name in names:
            if name in self._parameters:
                result.append(self._parameters[name])
        return result

    def get_parameters_by_path(self, path, with_decryption, recursive, filters=None):
        """Implement the get-parameters-by-path-API in the backend."""
        result = []
        # path could be with or without a trailing /. we handle this
        # difference here.
        path = path.rstrip('/') + '/'
        for param in self._parameters:
            if path != '/' and not param.startswith(path):
                continue
            if '/' in param[len(path) + 1:] and not recursive:
                continue
            if not self._match_filters(self._parameters[param], filters):
                continue
            result.append(self._parameters[param])

        return result

    @staticmethod
    def _match_filters(parameter, filters=None):
        """Return True if the given parameter matches all the filters"""
        for filter_obj in (filters or []):
            key = filter_obj['Key']
            option = filter_obj.get('Option', 'Equals')
            values = filter_obj.get('Values', [])

            what = None
            if key == 'Type':
                what = parameter.type
            elif key == 'KeyId':
                what = parameter.keyid

            if option == 'Equals'\
                    and not any(what == value for value in values):
                return False
            elif option == 'BeginsWith'\
                    and not any(what.startswith(value) for value in values):
                return False
        # True if no false match (or no filters at all)
        return True

    def get_parameter(self, name, with_decryption):
        if name in self._parameters:
            return self._parameters[name]
        return None

    def put_parameter(self, name, description, value, type, allowed_pattern,
            keyid, overwrite):
        previous_parameter = self._parameters.get(name)
        version = 1

        if previous_parameter:
            version = previous_parameter.version + 1

            if not overwrite:
                return

        last_modified_date = time.time()
        self._parameters[name] = Parameter(name, value, type, description,
            allowed_pattern, keyid, last_modified_date, version)
        return version

    def add_tags_to_resource(self, resource_type, resource_id, tags):
        for key, value in tags.items():
            self._resource_tags[resource_type][resource_id][key] = value

    def remove_tags_from_resource(self, resource_type, resource_id, keys):
        tags = self._resource_tags[resource_type][resource_id]
        for key in keys:
            if key in tags:
                del tags[key]

    def list_tags_for_resource(self, resource_type, resource_id):
        return self._resource_tags[resource_type][resource_id]

    def send_command(self, **kwargs):
        command = Command(
            comment=kwargs.get('Comment', ''),
            document_name=kwargs.get('DocumentName'),
            timeout_seconds=kwargs.get('TimeoutSeconds', 3600),
            instance_ids=kwargs.get('InstanceIds', []),
            max_concurrency=kwargs.get('MaxConcurrency', '50'),
            max_errors=kwargs.get('MaxErrors', '0'),
            notification_config=kwargs.get('NotificationConfig', {
                'NotificationArn': 'string',
                'NotificationEvents': ['Success'],
                'NotificationType': 'Command'
            }),
            output_s3_bucket_name=kwargs.get('OutputS3BucketName', ''),
            output_s3_key_prefix=kwargs.get('OutputS3KeyPrefix', ''),
            output_s3_region=kwargs.get('OutputS3Region', ''),
            parameters=kwargs.get('Parameters', {}),
            service_role_arn=kwargs.get('ServiceRoleArn', ''),
            targets=kwargs.get('Targets', []),
            backend_region=self._region)

        self._commands.append(command)
        return {
            'Command': command.response_object()
        }

    def list_commands(self, **kwargs):
        """
        https://docs.aws.amazon.com/systems-manager/latest/APIReference/API_ListCommands.html
        """
        commands = self._commands

        command_id = kwargs.get('CommandId', None)
        if command_id:
            commands = [self.get_command_by_id(command_id)]
        instance_id = kwargs.get('InstanceId', None)
        if instance_id:
            commands = self.get_commands_by_instance_id(instance_id)

        return {
            'Commands': [command.response_object() for command in commands]
        }

    def get_command_by_id(self, id):
        command = next(
            (command for command in self._commands if command.command_id == id), None)

        if command is None:
            raise RESTError('InvalidCommandId', 'Invalid command id.')

        return command

    def get_commands_by_instance_id(self, instance_id):
        return [
            command for command in self._commands
            if instance_id in command.instance_ids]

    def get_command_invocation(self, **kwargs):
        """
        https://docs.aws.amazon.com/systems-manager/latest/APIReference/API_GetCommandInvocation.html
        """

        command_id = kwargs.get('CommandId')
        instance_id = kwargs.get('InstanceId')
        plugin_name = kwargs.get('PluginName', None)

        command = self.get_command_by_id(command_id)
        return command.get_invocation(instance_id, plugin_name)


ssm_backends = {}
for region, ec2_backend in ec2_backends.items():
    ssm_backends[region] = SimpleSystemManagerBackend()



from __future__ import unicode_literals
from .responses import GlacierResponse

url_bases = [
    "https?://glacier.(.+).amazonaws.com",
]

url_paths = {
    '{0}/(?P<account_number>.+)/vaults$': GlacierResponse.all_vault_response,
    '{0}/(?P<account_number>.+)/vaults/(?P<vault_name>[^/.]+)$': GlacierResponse.vault_response,
    '{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/archives$': GlacierResponse.vault_archive_response,
    '{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/archives/(?P<archive_id>.+)$': GlacierResponse.vault_archive_individual_response,
    '{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/jobs$': GlacierResponse.vault_jobs_response,
    '{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/jobs/(?P<job_id>[^/.]+)$': GlacierResponse.vault_jobs_individual_response,
    '{0}/(?P<account_number>.+)/vaults/(?P<vault_name>.+)/jobs/(?P<job_id>.+)/output$': GlacierResponse.vault_jobs_output_response,
}



from __future__ import unicode_literals

import json
import re

from six.moves.urllib.parse import urlsplit

from moto.core.responses import BaseResponse
from .models import polly_backends
from .resources import LANGUAGE_CODES, VOICE_IDS

LEXICON_NAME_REGEX = re.compile(r'^[0-9A-Za-z]{1,20}$')


class PollyResponse(BaseResponse):
    @property
    def polly_backend(self):
        return polly_backends[self.region]

    @property
    def json(self):
        if not hasattr(self, '_json'):
            self._json = json.loads(self.body)
        return self._json

    def _error(self, code, message):
        return json.dumps({'__type': code, 'message': message}), dict(status=400)

    def _get_action(self):
        # Amazon is now naming things /v1/api_name
        url_parts = urlsplit(self.uri).path.lstrip('/').split('/')
        # [0] = 'v1'

        return url_parts[1]

    # DescribeVoices
    def voices(self):
        language_code = self._get_param('LanguageCode')
        next_token = self._get_param('NextToken')

        if language_code is not None and language_code not in LANGUAGE_CODES:
            msg = "1 validation error detected: Value '{0}' at 'languageCode' failed to satisfy constraint: " \
                  "Member must satisfy enum value set: [{1}]".format(language_code, ', '.join(LANGUAGE_CODES))
            return msg, dict(status=400)

        voices = self.polly_backend.describe_voices(language_code, next_token)

        return json.dumps({'Voices': voices})

    def lexicons(self):
        # Dish out requests based on methods

        # anything after the /v1/lexicons/
        args = urlsplit(self.uri).path.lstrip('/').split('/')[2:]

        if self.method == 'GET':
            if len(args) == 0:
                return self._get_lexicons_list()
            else:
                return self._get_lexicon(*args)
        elif self.method == 'PUT':
            return self._put_lexicons(*args)
        elif self.method == 'DELETE':
            return self._delete_lexicon(*args)

        return self._error('InvalidAction', 'Bad route')

    # PutLexicon
    def _put_lexicons(self, lexicon_name):
        if LEXICON_NAME_REGEX.match(lexicon_name) is None:
            return self._error('InvalidParameterValue', 'Lexicon name must match [0-9A-Za-z]{1,20}')

        if 'Content' not in self.json:
            return self._error('MissingParameter', 'Content is missing from the body')

        self.polly_backend.put_lexicon(lexicon_name, self.json['Content'])

        return ''

    # ListLexicons
    def _get_lexicons_list(self):
        next_token = self._get_param('NextToken')

        result = {
            'Lexicons': self.polly_backend.list_lexicons(next_token)
        }

        return json.dumps(result)

    # GetLexicon
    def _get_lexicon(self, lexicon_name):
        try:
            lexicon = self.polly_backend.get_lexicon(lexicon_name)
        except KeyError:
            return self._error('LexiconNotFoundException', 'Lexicon not found')

        result = {
            'Lexicon': {
                'Name': lexicon_name,
                'Content': lexicon.content
            },
            'LexiconAttributes': lexicon.to_dict()['Attributes']
        }

        return json.dumps(result)

    # DeleteLexicon
    def _delete_lexicon(self, lexicon_name):
        try:
            self.polly_backend.delete_lexicon(lexicon_name)
        except KeyError:
            return self._error('LexiconNotFoundException', 'Lexicon not found')

        return ''

    # SynthesizeSpeech
    def speech(self):
        # Sanity check params
        args = {
            'lexicon_names': None,
            'sample_rate': 22050,
            'speech_marks': None,
            'text': None,
            'text_type': 'text'
        }

        if 'LexiconNames' in self.json:
            for lex in self.json['LexiconNames']:
                try:
                    self.polly_backend.get_lexicon(lex)
                except KeyError:
                    return self._error('LexiconNotFoundException', 'Lexicon not found')

            args['lexicon_names'] = self.json['LexiconNames']

        if 'OutputFormat' not in self.json:
            return self._error('MissingParameter', 'Missing parameter OutputFormat')
        if self.json['OutputFormat'] not in ('json', 'mp3', 'ogg_vorbis', 'pcm'):
            return self._error('InvalidParameterValue', 'Not one of json, mp3, ogg_vorbis, pcm')
        args['output_format'] = self.json['OutputFormat']

        if 'SampleRate' in self.json:
            sample_rate = int(self.json['SampleRate'])
            if sample_rate not in (8000, 16000, 22050):
                return self._error('InvalidSampleRateException', 'The specified sample rate is not valid.')
            args['sample_rate'] = sample_rate

        if 'SpeechMarkTypes' in self.json:
            for value in self.json['SpeechMarkTypes']:
                if value not in ('sentance', 'ssml', 'viseme', 'word'):
                    return self._error('InvalidParameterValue', 'Not one of sentance, ssml, viseme, word')
            args['speech_marks'] = self.json['SpeechMarkTypes']

        if 'Text' not in self.json:
            return self._error('MissingParameter', 'Missing parameter Text')
        args['text'] = self.json['Text']

        if 'TextType' in self.json:
            if self.json['TextType'] not in ('ssml', 'text'):
                return self._error('InvalidParameterValue', 'Not one of ssml, text')
            args['text_type'] = self.json['TextType']

        if 'VoiceId' not in self.json:
            return self._error('MissingParameter', 'Missing parameter VoiceId')
        if self.json['VoiceId'] not in VOICE_IDS:
            return self._error('InvalidParameterValue', 'Not one of {0}'.format(', '.join(VOICE_IDS)))
        args['voice_id'] = self.json['VoiceId']

        # More validation
        if len(args['text']) > 3000:
            return self._error('TextLengthExceededException', 'Text too long')

        if args['speech_marks'] is not None and args['output_format'] != 'json':
            return self._error('MarksNotSupportedForFormatException', 'OutputFormat must be json')
        if args['speech_marks'] is not None and args['text_type'] == 'text':
            return self._error('SsmlMarksNotSupportedForTextTypeException', 'TextType must be ssml')

        content_type = 'audio/json'
        if args['output_format'] == 'mp3':
            content_type = 'audio/mpeg'
        elif args['output_format'] == 'ogg_vorbis':
            content_type = 'audio/ogg'
        elif args['output_format'] == 'pcm':
            content_type = 'audio/pcm'

        headers = {'Content-Type': content_type}

        return '\x00\x00\x00\x00\x00\x00\x00\x00', headers



from __future__ import unicode_literals

import json
import yaml
from six.moves.urllib.parse import urlparse

from moto.core.responses import BaseResponse
from moto.core.utils import amzn_request_id
from moto.s3 import s3_backend
from .models import cloudformation_backends
from .exceptions import ValidationError


class CloudFormationResponse(BaseResponse):

    @property
    def cloudformation_backend(self):
        return cloudformation_backends[self.region]

    def _get_stack_from_s3_url(self, template_url):
        template_url_parts = urlparse(template_url)
        if "localhost" in template_url:
            bucket_name, key_name = template_url_parts.path.lstrip(
                "/").split("/", 1)
        else:
            if template_url_parts.netloc.endswith('amazonaws.com') \
                    and template_url_parts.netloc.startswith('s3'):
                # Handle when S3 url uses amazon url with bucket in path
                # Also handles getting region as technically s3 is region'd

                # region = template_url.netloc.split('.')[1]
                bucket_name, key_name = template_url_parts.path.lstrip(
                    "/").split("/", 1)
            else:
                bucket_name = template_url_parts.netloc.split(".")[0]
                key_name = template_url_parts.path.lstrip("/")

        key = s3_backend.get_key(bucket_name, key_name)
        return key.value.decode("utf-8")

    def create_stack(self):
        stack_name = self._get_param('StackName')
        stack_body = self._get_param('TemplateBody')
        template_url = self._get_param('TemplateURL')
        role_arn = self._get_param('RoleARN')
        parameters_list = self._get_list_prefix("Parameters.member")
        tags = dict((item['key'], item['value'])
                    for item in self._get_list_prefix("Tags.member"))

        # Hack dict-comprehension
        parameters = dict([
            (parameter['parameter_key'], parameter['parameter_value'])
            for parameter
            in parameters_list
        ])
        if template_url:
            stack_body = self._get_stack_from_s3_url(template_url)
        stack_notification_arns = self._get_multi_param(
            'NotificationARNs.member')

        stack = self.cloudformation_backend.create_stack(
            name=stack_name,
            template=stack_body,
            parameters=parameters,
            region_name=self.region,
            notification_arns=stack_notification_arns,
            tags=tags,
            role_arn=role_arn,
        )
        if self.request_json:
            return json.dumps({
                'CreateStackResponse': {
                    'CreateStackResult': {
                        'StackId': stack.stack_id,
                    }
                }
            })
        else:
            template = self.response_template(CREATE_STACK_RESPONSE_TEMPLATE)
            return template.render(stack=stack)

    @amzn_request_id
    def create_change_set(self):
        stack_name = self._get_param('StackName')
        change_set_name = self._get_param('ChangeSetName')
        stack_body = self._get_param('TemplateBody')
        template_url = self._get_param('TemplateURL')
        role_arn = self._get_param('RoleARN')
        update_or_create = self._get_param('ChangeSetType', 'CREATE')
        parameters_list = self._get_list_prefix("Parameters.member")
        tags = dict((item['key'], item['value'])
                    for item in self._get_list_prefix("Tags.member"))
        parameters = {param['parameter_key']: param['parameter_value']
                      for param in parameters_list}
        if template_url:
            stack_body = self._get_stack_from_s3_url(template_url)
        stack_notification_arns = self._get_multi_param(
            'NotificationARNs.member')
        change_set_id, stack_id = self.cloudformation_backend.create_change_set(
            stack_name=stack_name,
            change_set_name=change_set_name,
            template=stack_body,
            parameters=parameters,
            region_name=self.region,
            notification_arns=stack_notification_arns,
            tags=tags,
            role_arn=role_arn,
            change_set_type=update_or_create,
        )
        if self.request_json:
            return json.dumps({
                'CreateChangeSetResponse': {
                    'CreateChangeSetResult': {
                        'Id': change_set_id,
                        'StackId': stack_id,
                    }
                }
            })
        else:
            template = self.response_template(CREATE_CHANGE_SET_RESPONSE_TEMPLATE)
            return template.render(stack_id=stack_id, change_set_id=change_set_id)

    def delete_change_set(self):
        stack_name = self._get_param('StackName')
        change_set_name = self._get_param('ChangeSetName')

        self.cloudformation_backend.delete_change_set(change_set_name=change_set_name, stack_name=stack_name)
        if self.request_json:
            return json.dumps({
                'DeleteChangeSetResponse': {
                    'DeleteChangeSetResult': {},
                }
            })
        else:
            template = self.response_template(DELETE_CHANGE_SET_RESPONSE_TEMPLATE)
            return template.render()

    def describe_change_set(self):
        stack_name = self._get_param('StackName')
        change_set_name = self._get_param('ChangeSetName')
        change_set = self.cloudformation_backend.describe_change_set(
            change_set_name=change_set_name,
            stack_name=stack_name,
        )
        template = self.response_template(DESCRIBE_CHANGE_SET_RESPONSE_TEMPLATE)
        return template.render(change_set=change_set)

    @amzn_request_id
    def execute_change_set(self):
        stack_name = self._get_param('StackName')
        change_set_name = self._get_param('ChangeSetName')
        self.cloudformation_backend.execute_change_set(
            stack_name=stack_name,
            change_set_name=change_set_name,
        )
        if self.request_json:
            return json.dumps({
                'ExecuteChangeSetResponse': {
                    'ExecuteChangeSetResult': {},
                }
            })
        else:
            template = self.response_template(EXECUTE_CHANGE_SET_RESPONSE_TEMPLATE)
            return template.render()

    def describe_stacks(self):
        stack_name_or_id = None
        if self._get_param('StackName'):
            stack_name_or_id = self.querystring.get('StackName')[0]
        token = self._get_param('NextToken')
        stacks = self.cloudformation_backend.describe_stacks(stack_name_or_id)
        stack_ids = [stack.stack_id for stack in stacks]
        if token:
            start = stack_ids.index(token) + 1
        else:
            start = 0
        max_results = 50  # using this to mske testing of paginated stacks more convenient than default 1 MB
        stacks_resp = stacks[start:start + max_results]
        next_token = None
        if len(stacks) > (start + max_results):
            next_token = stacks_resp[-1].stack_id
        template = self.response_template(DESCRIBE_STACKS_TEMPLATE)
        return template.render(stacks=stacks_resp, next_token=next_token)

    def describe_stack_resource(self):
        stack_name = self._get_param('StackName')
        stack = self.cloudformation_backend.get_stack(stack_name)
        logical_resource_id = self._get_param('LogicalResourceId')

        for stack_resource in stack.stack_resources:
            if stack_resource.logical_resource_id == logical_resource_id:
                resource = stack_resource
                break
        else:
            raise ValidationError(logical_resource_id)

        template = self.response_template(
            DESCRIBE_STACK_RESOURCE_RESPONSE_TEMPLATE)
        return template.render(stack=stack, resource=resource)

    def describe_stack_resources(self):
        stack_name = self._get_param('StackName')
        stack = self.cloudformation_backend.get_stack(stack_name)

        template = self.response_template(DESCRIBE_STACK_RESOURCES_RESPONSE)
        return template.render(stack=stack)

    def describe_stack_events(self):
        stack_name = self._get_param('StackName')
        stack = self.cloudformation_backend.get_stack(stack_name)

        template = self.response_template(DESCRIBE_STACK_EVENTS_RESPONSE)
        return template.render(stack=stack)

    def list_change_sets(self):
        change_sets = self.cloudformation_backend.list_change_sets()
        template = self.response_template(LIST_CHANGE_SETS_RESPONSE)
        return template.render(change_sets=change_sets)

    def list_stacks(self):
        stacks = self.cloudformation_backend.list_stacks()
        template = self.response_template(LIST_STACKS_RESPONSE)
        return template.render(stacks=stacks)

    def list_stack_resources(self):
        stack_name_or_id = self._get_param('StackName')
        resources = self.cloudformation_backend.list_stack_resources(
            stack_name_or_id)

        template = self.response_template(LIST_STACKS_RESOURCES_RESPONSE)
        return template.render(resources=resources)

    def get_template(self):
        name_or_stack_id = self.querystring.get('StackName')[0]
        stack = self.cloudformation_backend.get_stack(name_or_stack_id)

        if self.request_json:
            return json.dumps({
                "GetTemplateResponse": {
                    "GetTemplateResult": {
                        "TemplateBody": stack.template,
                        "ResponseMetadata": {
                            "RequestId": "2d06e36c-ac1d-11e0-a958-f9382b6eb86bEXAMPLE"
                        }
                    }
                }
            })
        else:
            template = self.response_template(GET_TEMPLATE_RESPONSE_TEMPLATE)
            return template.render(stack=stack)

    def update_stack(self):
        stack_name = self._get_param('StackName')
        role_arn = self._get_param('RoleARN')
        template_url = self._get_param('TemplateURL')
        stack_body = self._get_param('TemplateBody')
        stack = self.cloudformation_backend.get_stack(stack_name)
        if self._get_param('UsePreviousTemplate') == "true":
            stack_body = stack.template
        elif not stack_body and template_url:
            stack_body = self._get_stack_from_s3_url(template_url)

        incoming_params = self._get_list_prefix("Parameters.member")
        parameters = dict([
            (parameter['parameter_key'], parameter['parameter_value'])
            for parameter
            in incoming_params if 'parameter_value' in parameter
        ])
        previous = dict([
            (parameter['parameter_key'], stack.parameters[parameter['parameter_key']])
            for parameter
            in incoming_params if 'use_previous_value' in parameter
        ])
        parameters.update(previous)
        # boto3 is supposed to let you clear the tags by passing an empty value, but the request body doesn't
        # end up containing anything we can use to differentiate between passing an empty value versus not
        # passing anything. so until that changes, moto won't be able to clear tags, only update them.
        tags = dict((item['key'], item['value'])
                    for item in self._get_list_prefix("Tags.member"))
        # so that if we don't pass the parameter, we don't clear all the tags accidentally
        if not tags:
            tags = None

        stack = self.cloudformation_backend.get_stack(stack_name)
        if stack.status == 'ROLLBACK_COMPLETE':
            raise ValidationError(
                stack.stack_id, message="Stack:{0} is in ROLLBACK_COMPLETE state and can not be updated.".format(stack.stack_id))

        stack = self.cloudformation_backend.update_stack(
            name=stack_name,
            template=stack_body,
            role_arn=role_arn,
            parameters=parameters,
            tags=tags,
        )
        if self.request_json:
            stack_body = {
                'UpdateStackResponse': {
                    'UpdateStackResult': {
                        'StackId': stack.name,
                    }
                }
            }
            return json.dumps(stack_body)
        else:
            template = self.response_template(UPDATE_STACK_RESPONSE_TEMPLATE)
            return template.render(stack=stack)

    def delete_stack(self):
        name_or_stack_id = self.querystring.get('StackName')[0]

        self.cloudformation_backend.delete_stack(name_or_stack_id)
        if self.request_json:
            return json.dumps({
                'DeleteStackResponse': {
                    'DeleteStackResult': {},
                }
            })
        else:
            template = self.response_template(DELETE_STACK_RESPONSE_TEMPLATE)
            return template.render()

    def list_exports(self):
        token = self._get_param('NextToken')
        exports, next_token = self.cloudformation_backend.list_exports(token=token)
        template = self.response_template(LIST_EXPORTS_RESPONSE)
        return template.render(exports=exports, next_token=next_token)

    def validate_template(self):
        cfn_lint = self.cloudformation_backend.validate_template(self._get_param('TemplateBody'))
        if cfn_lint:
            raise ValidationError(cfn_lint[0].message)
        description = ""
        try:
            description = json.loads(self._get_param('TemplateBody'))['Description']
        except (ValueError, KeyError):
            pass
        try:
            description = yaml.load(self._get_param('TemplateBody'))['Description']
        except (yaml.ParserError, KeyError):
            pass
        template = self.response_template(VALIDATE_STACK_RESPONSE_TEMPLATE)
        return template.render(description=description)

    def create_stack_set(self):
        stackset_name = self._get_param('StackSetName')
        stack_body = self._get_param('TemplateBody')
        template_url = self._get_param('TemplateURL')
        # role_arn = self._get_param('RoleARN')
        parameters_list = self._get_list_prefix("Parameters.member")
        tags = dict((item['key'], item['value'])
                    for item in self._get_list_prefix("Tags.member"))

        # Copy-Pasta - Hack dict-comprehension
        parameters = dict([
            (parameter['parameter_key'], parameter['parameter_value'])
            for parameter
            in parameters_list
        ])
        if template_url:
            stack_body = self._get_stack_from_s3_url(template_url)

        stackset = self.cloudformation_backend.create_stack_set(
            name=stackset_name,
            template=stack_body,
            parameters=parameters,
            tags=tags,
            # role_arn=role_arn,
        )
        if self.request_json:
            return json.dumps({
                'CreateStackSetResponse': {
                    'CreateStackSetResult': {
                        'StackSetId': stackset.stackset_id,
                    }
                }
            })
        else:
            template = self.response_template(CREATE_STACK_SET_RESPONSE_TEMPLATE)
            return template.render(stackset=stackset)

    def create_stack_instances(self):
        stackset_name = self._get_param('StackSetName')
        accounts = self._get_multi_param('Accounts.member')
        regions = self._get_multi_param('Regions.member')
        parameters = self._get_multi_param('ParameterOverrides.member')
        self.cloudformation_backend.create_stack_instances(stackset_name, accounts, regions, parameters)
        template = self.response_template(CREATE_STACK_INSTANCES_TEMPLATE)
        return template.render()

    def delete_stack_set(self):
        stackset_name = self._get_param('StackSetName')
        self.cloudformation_backend.delete_stack_set(stackset_name)
        template = self.response_template(DELETE_STACK_SET_RESPONSE_TEMPLATE)
        return template.render()

    def delete_stack_instances(self):
        stackset_name = self._get_param('StackSetName')
        accounts = self._get_multi_param('Accounts.member')
        regions = self._get_multi_param('Regions.member')
        operation = self.cloudformation_backend.delete_stack_instances(stackset_name, accounts, regions)

        template = self.response_template(DELETE_STACK_INSTANCES_TEMPLATE)
        return template.render(operation=operation)

    def describe_stack_set(self):
        stackset_name = self._get_param('StackSetName')
        stackset = self.cloudformation_backend.get_stack_set(stackset_name)

        if not stackset.admin_role:
            stackset.admin_role = 'arn:aws:iam::123456789012:role/AWSCloudFormationStackSetAdministrationRole'
        if not stackset.execution_role:
            stackset.execution_role = 'AWSCloudFormationStackSetExecutionRole'

        template = self.response_template(DESCRIBE_STACK_SET_RESPONSE_TEMPLATE)
        return template.render(stackset=stackset)

    def describe_stack_instance(self):
        stackset_name = self._get_param('StackSetName')
        account = self._get_param('StackInstanceAccount')
        region = self._get_param('StackInstanceRegion')

        instance = self.cloudformation_backend.get_stack_set(stackset_name).instances.get_instance(account, region)
        template = self.response_template(DESCRIBE_STACK_INSTANCE_TEMPLATE)
        rendered = template.render(instance=instance)
        return rendered

    def list_stack_sets(self):
        stacksets = self.cloudformation_backend.stacksets
        template = self.response_template(LIST_STACK_SETS_TEMPLATE)
        return template.render(stacksets=stacksets)

    def list_stack_instances(self):
        stackset_name = self._get_param('StackSetName')
        stackset = self.cloudformation_backend.get_stack_set(stackset_name)
        template = self.response_template(LIST_STACK_INSTANCES_TEMPLATE)
        return template.render(stackset=stackset)

    def list_stack_set_operations(self):
        stackset_name = self._get_param('StackSetName')
        stackset = self.cloudformation_backend.get_stack_set(stackset_name)
        template = self.response_template(LIST_STACK_SET_OPERATIONS_RESPONSE_TEMPLATE)
        return template.render(stackset=stackset)

    def stop_stack_set_operation(self):
        stackset_name = self._get_param('StackSetName')
        operation_id = self._get_param('OperationId')
        stackset = self.cloudformation_backend.get_stack_set(stackset_name)
        stackset.update_operation(operation_id, 'STOPPED')
        template = self.response_template(STOP_STACK_SET_OPERATION_RESPONSE_TEMPLATE)
        return template.render()

    def describe_stack_set_operation(self):
        stackset_name = self._get_param('StackSetName')
        operation_id = self._get_param('OperationId')
        stackset = self.cloudformation_backend.get_stack_set(stackset_name)
        operation = stackset.get_operation(operation_id)
        template = self.response_template(DESCRIBE_STACKSET_OPERATION_RESPONSE_TEMPLATE)
        return template.render(stackset=stackset, operation=operation)

    def list_stack_set_operation_results(self):
        stackset_name = self._get_param('StackSetName')
        operation_id = self._get_param('OperationId')
        stackset = self.cloudformation_backend.get_stack_set(stackset_name)
        operation = stackset.get_operation(operation_id)
        template = self.response_template(LIST_STACK_SET_OPERATION_RESULTS_RESPONSE_TEMPLATE)
        return template.render(operation=operation)

    def update_stack_set(self):
        stackset_name = self._get_param('StackSetName')
        operation_id = self._get_param('OperationId')
        description = self._get_param('Description')
        execution_role = self._get_param('ExecutionRoleName')
        admin_role = self._get_param('AdministrationRoleARN')
        accounts = self._get_multi_param('Accounts.member')
        regions = self._get_multi_param('Regions.member')
        template_body = self._get_param('TemplateBody')
        template_url = self._get_param('TemplateURL')
        if template_url:
            template_body = self._get_stack_from_s3_url(template_url)
        tags = dict((item['key'], item['value'])
                    for item in self._get_list_prefix("Tags.member"))
        parameters_list = self._get_list_prefix("Parameters.member")
        parameters = dict([
            (parameter['parameter_key'], parameter['parameter_value'])
            for parameter
            in parameters_list
        ])
        operation = self.cloudformation_backend.update_stack_set(
            stackset_name=stackset_name,
            template=template_body,
            description=description,
            parameters=parameters,
            tags=tags,
            admin_role=admin_role,
            execution_role=execution_role,
            accounts=accounts,
            regions=regions,
            operation_id=operation_id
        )

        template = self.response_template(UPDATE_STACK_SET_RESPONSE_TEMPLATE)
        return template.render(operation=operation)

    def update_stack_instances(self):
        stackset_name = self._get_param('StackSetName')
        accounts = self._get_multi_param('Accounts.member')
        regions = self._get_multi_param('Regions.member')
        parameters = self._get_multi_param('ParameterOverrides.member')
        operation = self.cloudformation_backend.get_stack_set(stackset_name).update_instances(accounts, regions, parameters)
        template = self.response_template(UPDATE_STACK_INSTANCES_RESPONSE_TEMPLATE)
        return template.render(operation=operation)


VALIDATE_STACK_RESPONSE_TEMPLATE = """<ValidateTemplateResponse>
        <ValidateTemplateResult>
        <Capabilities></Capabilities>
<CapabilitiesReason></CapabilitiesReason>
<DeclaredTransforms></DeclaredTransforms>
<Description>{{ description }}</Description>
<Parameters></Parameters>
</ValidateTemplateResult>
</ValidateTemplateResponse>"""

CREATE_STACK_RESPONSE_TEMPLATE = """<CreateStackResponse>
  <CreateStackResult>
    <StackId>{{ stack.stack_id }}</StackId>
  </CreateStackResult>
  <ResponseMetadata>
    <RequestId>b9b4b068-3a41-11e5-94eb-example</RequestId>
    </ResponseMetadata>
</CreateStackResponse>
"""

UPDATE_STACK_RESPONSE_TEMPLATE = """<UpdateStackResponse xmlns="http://cloudformation.amazonaws.com/doc/2010-05-15/">
  <UpdateStackResult>
    <StackId>{{ stack.stack_id }}</StackId>
  </UpdateStackResult>
  <ResponseMetadata>
    <RequestId>b9b4b068-3a41-11e5-94eb-example</RequestId>
  </ResponseMetadata>
</UpdateStackResponse>
"""

CREATE_CHANGE_SET_RESPONSE_TEMPLATE = """<CreateStackResponse>
  <CreateChangeSetResult>
    <Id>{{change_set_id}}</Id>
    <StackId>{{ stack_id }}</StackId>
  </CreateChangeSetResult>
  <ResponseMetadata>
    <RequestId>{{ request_id }}</RequestId>
  </ResponseMetadata>
</CreateStackResponse>
"""

DELETE_CHANGE_SET_RESPONSE_TEMPLATE = """<DeleteChangeSetResponse>
  <DeleteChangeSetResult>
  </DeleteChangeSetResult>
  <ResponseMetadata>
    <RequestId>3d3200a1-810e-3023-6cc3-example</RequestId>
  </ResponseMetadata>
</DeleteChangeSetResponse>
"""

DESCRIBE_CHANGE_SET_RESPONSE_TEMPLATE = """<DescribeChangeSetResponse>
  <DescribeChangeSetResult>
    <ChangeSetId>{{ change_set.change_set_id }}</ChangeSetId>
    <ChangeSetName>{{ change_set.change_set_name }}</ChangeSetName>
    <StackId>{{ change_set.stack_id }}</StackId>
    <StackName>{{ change_set.stack_name }}</StackName>
    <Description>{{ change_set.description }}</Description>
    <Parameters>
      {% for param_name, param_value in change_set.stack_parameters.items() %}
       <member>
          <ParameterKey>{{ param_name }}</ParameterKey>
          <ParameterValue>{{ param_value }}</ParameterValue>
        </member>
      {% endfor %}
    </Parameters>
    <CreationTime>2011-05-23T15:47:44Z</CreationTime>
    <ExecutionStatus>{{ change_set.execution_status }}</ExecutionStatus>
    <Status>{{ change_set.status }}</Status>
    <StatusReason>{{ change_set.status_reason }}</StatusReason>
    {% if change_set.notification_arns %}
    <NotificationARNs>
      {% for notification_arn in change_set.notification_arns %}
      <member>{{ notification_arn }}</member>
      {% endfor %}
    </NotificationARNs>
    {% else %}
    <NotificationARNs/>
    {% endif %}
    {% if change_set.role_arn %}
    <RoleARN>{{ change_set.role_arn }}</RoleARN>
    {% endif %}
    {% if change_set.changes %}
    <Changes>
      {% for change in change_set.changes %}
      <member>
        <Type>Resource</Type>
        <ResourceChange>
            <Action>{{ change.action }}</Action>
            <LogicalResourceId>{{ change.logical_resource_id }}</LogicalResourceId>
            <ResourceType>{{ change.resource_type }}</ResourceType>
        </ResourceChange>
      </member>
      {% endfor %}
    </Changes>
    {% endif %}
    {% if next_token %}
    <NextToken>{{ next_token }}</NextToken>
    {% endif %}
  </DescribeChangeSetResult>
</DescribeChangeSetResponse>"""

EXECUTE_CHANGE_SET_RESPONSE_TEMPLATE = """<ExecuteChangeSetResponse>
  <ExecuteChangeSetResult>
      <ExecuteChangeSetResult/>
  </ExecuteChangeSetResult>
  <ResponseMetadata>
    <RequestId>{{ request_id }}</RequestId>
  </ResponseMetadata>
</ExecuteChangeSetResponse>
"""

DESCRIBE_STACKS_TEMPLATE = """<DescribeStacksResponse>
  <DescribeStacksResult>
    <Stacks>
      {% for stack in stacks %}
      <member>
        <StackName>{{ stack.name }}</StackName>
        <StackId>{{ stack.stack_id }}</StackId>
        <CreationTime>2010-07-27T22:28:28Z</CreationTime>
        <StackStatus>{{ stack.status }}</StackStatus>
        {% if stack.notification_arns %}
        <NotificationARNs>
          {% for notification_arn in stack.notification_arns %}
          <member>{{ notification_arn }}</member>
          {% endfor %}
        </NotificationARNs>
        {% else %}
        <NotificationARNs/>
        {% endif %}
        <DisableRollback>false</DisableRollback>
        <Outputs>
        {% for output in stack.stack_outputs %}
          <member>
            <OutputKey>{{ output.key }}</OutputKey>
            <OutputValue>{{ output.value }}</OutputValue>
          </member>
        {% endfor %}
        </Outputs>
        <Parameters>
        {% for param_name, param_value in stack.stack_parameters.items() %}
          <member>
            <ParameterKey>{{ param_name }}</ParameterKey>
            {% if param_name in stack.resource_map.no_echo_parameter_keys %}
                <ParameterValue>****</ParameterValue>
            {% else %}
                <ParameterValue>{{ param_value }}</ParameterValue>
            {% endif %}
          </member>
        {% endfor %}
        </Parameters>
        {% if stack.role_arn %}
        <RoleARN>{{ stack.role_arn }}</RoleARN>
        {% endif %}
        <Tags>
          {% for tag_key, tag_value in stack.tags.items() %}
            <member>
              <Key>{{ tag_key }}</Key>
              <Value>{{ tag_value }}</Value>
            </member>
          {% endfor %}
        </Tags>
      </member>
      {% endfor %}
    </Stacks>
    {% if next_token %}
    <NextToken>{{ next_token }}</NextToken>
    {% endif %}
  </DescribeStacksResult>
</DescribeStacksResponse>"""


DESCRIBE_STACK_RESOURCE_RESPONSE_TEMPLATE = """<DescribeStackResourceResponse>
  <DescribeStackResourceResult>
    <StackResourceDetail>
      <StackId>{{ stack.stack_id }}</StackId>
      <StackName>{{ stack.name }}</StackName>
      <LogicalResourceId>{{ resource.logical_resource_id }}</LogicalResourceId>
      <PhysicalResourceId>{{ resource.physical_resource_id }}</PhysicalResourceId>
      <ResourceType>{{ resource.type }}</ResourceType>
      <Timestamp>2010-07-27T22:27:28Z</Timestamp>
      <ResourceStatus>{{ stack.status }}</ResourceStatus>
    </StackResourceDetail>
  </DescribeStackResourceResult>
</DescribeStackResourceResponse>"""


DESCRIBE_STACK_RESOURCES_RESPONSE = """<DescribeStackResourcesResponse>
    <DescribeStackResourcesResult>
      <StackResources>
        {% for resource in stack.stack_resources %}
        <member>
          <StackId>{{ stack.stack_id }}</StackId>
          <StackName>{{ stack.name }}</StackName>
          <LogicalResourceId>{{ resource.logical_resource_id }}</LogicalResourceId>
          <PhysicalResourceId>{{ resource.physical_resource_id }}</PhysicalResourceId>
          <ResourceType>{{ resource.type }}</ResourceType>
          <Timestamp>2010-07-27T22:27:28Z</Timestamp>
          <ResourceStatus>{{ stack.status }}</ResourceStatus>
        </member>
        {% endfor %}
      </StackResources>
    </DescribeStackResourcesResult>
</DescribeStackResourcesResponse>"""


DESCRIBE_STACK_EVENTS_RESPONSE = """<DescribeStackEventsResponse xmlns="http://cloudformation.amazonaws.com/doc/2010-05-15/">
  <DescribeStackEventsResult>
    <StackEvents>
      {% for event in stack.events[::-1] %}
      <member>
        <Timestamp>{{ event.timestamp.strftime('%Y-%m-%dT%H:%M:%S.%fZ') }}</Timestamp>
        <ResourceStatus>{{ event.resource_status }}</ResourceStatus>
        <StackId>{{ event.stack_id }}</StackId>
        <EventId>{{ event.event_id }}</EventId>
        <LogicalResourceId>{{ event.logical_resource_id }}</LogicalResourceId>
        {% if event.resource_status_reason %}<ResourceStatusReason>{{ event.resource_status_reason }}</ResourceStatusReason>{% endif %}
        <StackName>{{ event.stack_name }}</StackName>
        <PhysicalResourceId>{{ event.physical_resource_id }}</PhysicalResourceId>
        {% if event.resource_properties %}<ResourceProperties>{{ event.resource_properties }}</ResourceProperties>{% endif %}
        <ResourceType>{{ event.resource_type }}</ResourceType>
      </member>
      {% endfor %}
    </StackEvents>
  </DescribeStackEventsResult>
  <ResponseMetadata>
    <RequestId>b9b4b068-3a41-11e5-94eb-example</RequestId>
  </ResponseMetadata>
</DescribeStackEventsResponse>"""


LIST_CHANGE_SETS_RESPONSE = """<ListChangeSetsResponse>
 <ListChangeSetsResult>
  <Summaries>
    {% for change_set in change_sets %}
    <member>
        <StackId>{{ change_set.stack_id }}</StackId>
        <StackName>{{ change_set.stack_name }}</StackName>
        <ChangeSetId>{{ change_set.change_set_id }}</ChangeSetId>
        <ChangeSetName>{{ change_set.change_set_name }}</ChangeSetName>
        <ExecutionStatus>{{ change_set.execution_status }}</ExecutionStatus>
        <Status>{{ change_set.status }}</Status>
        <StatusReason>{{ change_set.status_reason }}</StatusReason>
        <CreationTime>2011-05-23T15:47:44Z</CreationTime>
        <Description>{{ change_set.description }}</Description>
    </member>
    {% endfor %}
  </Summaries>
 </ListChangeSetsResult>
</ListChangeSetsResponse>"""


LIST_STACKS_RESPONSE = """<ListStacksResponse>
 <ListStacksResult>
  <StackSummaries>
    {% for stack in stacks %}
    <member>
        <StackId>{{ stack.stack_id }}</StackId>
        <StackStatus>{{ stack.status }}</StackStatus>
        <StackName>{{ stack.name }}</StackName>
        <CreationTime>2011-05-23T15:47:44Z</CreationTime>
        <TemplateDescription>{{ stack.description }}</TemplateDescription>
    </member>
    {% endfor %}
  </StackSummaries>
 </ListStacksResult>
</ListStacksResponse>"""


LIST_STACKS_RESOURCES_RESPONSE = """<ListStackResourcesResponse>
  <ListStackResourcesResult>
    <StackResourceSummaries>
      {% for resource in resources %}
      <member>
        <ResourceStatus>CREATE_COMPLETE</ResourceStatus>
        <LogicalResourceId>{{ resource.logical_resource_id }}</LogicalResourceId>
        <LastUpdatedTimestamp>2011-06-21T20:15:58Z</LastUpdatedTimestamp>
        <PhysicalResourceId>{{ resource.physical_resource_id }}</PhysicalResourceId>
        <ResourceType>{{ resource.type }}</ResourceType>
      </member>
      {% endfor %}
    </StackResourceSummaries>
  </ListStackResourcesResult>
  <ResponseMetadata>
    <RequestId>2d06e36c-ac1d-11e0-a958-f9382b6eb86b</RequestId>
  </ResponseMetadata>
</ListStackResourcesResponse>"""


GET_TEMPLATE_RESPONSE_TEMPLATE = """<GetTemplateResponse>
  <GetTemplateResult>
    <TemplateBody>{{ stack.template }}</TemplateBody>
  </GetTemplateResult>
  <ResponseMetadata>
    <RequestId>b9b4b068-3a41-11e5-94eb-example</RequestId>
  </ResponseMetadata>
</GetTemplateResponse>"""


DELETE_STACK_RESPONSE_TEMPLATE = """<DeleteStackResponse>
  <ResponseMetadata>
    <RequestId>5ccc7dcd-744c-11e5-be70-example</RequestId>
  </ResponseMetadata>
</DeleteStackResponse>
"""


LIST_EXPORTS_RESPONSE = """<ListExportsResponse xmlns="http://cloudformation.amazonaws.com/doc/2010-05-15/">
  <ListExportsResult>
    <Exports>
      {% for export in exports %}
      <member>
        <ExportingStackId>{{ export.exporting_stack_id }}</ExportingStackId>
        <Name>{{ export.name }}</Name>
        <Value>{{ export.value }}</Value>
      </member>
      {% endfor %}
    </Exports>
    {% if next_token %}
    <NextToken>{{ next_token }}</NextToken>
    {% endif %}
  </ListExportsResult>
  <ResponseMetadata>
    <RequestId>5ccc7dcd-744c-11e5-be70-example</RequestId>
  </ResponseMetadata>
</ListExportsResponse>"""

CREATE_STACK_SET_RESPONSE_TEMPLATE = """<CreateStackSetResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <CreateStackSetResult>
    <StackSetId>{{ stackset.stackset_id }}</StackSetId>
  </CreateStackSetResult>
  <ResponseMetadata>
    <RequestId>f457258c-391d-41d1-861f-example</RequestId>
  </ResponseMetadata>
</CreateStackSetResponse>
"""

DESCRIBE_STACK_SET_RESPONSE_TEMPLATE = """<DescribeStackSetResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <DescribeStackSetResult>
    <StackSet>
      <Capabilities/>
      <StackSetARN>{{ stackset.arn }}</StackSetARN>
      <ExecutionRoleName>{{ stackset.execution_role }}</ExecutionRoleName>
      <AdministrationRoleARN>{{ stackset.admin_role }}</AdministrationRoleARN>
      <StackSetId>{{ stackset.id }}</StackSetId>
      <TemplateBody>{{ stackset.template }}</TemplateBody>
      <StackSetName>{{ stackset.name }}</StackSetName>
        <Parameters>
        {% for param_name, param_value in stackset.parameters.items() %}
          <member>
            <ParameterKey>{{ param_name }}</ParameterKey>
            <ParameterValue>{{ param_value }}</ParameterValue>
          </member>
        {% endfor %}
        </Parameters>
        <Tags>
          {% for tag_key, tag_value in stackset.tags.items() %}
            <member>
              <Key>{{ tag_key }}</Key>
              <Value>{{ tag_value }}</Value>
            </member>
          {% endfor %}
        </Tags>
      <Status>{{ stackset.status }}</Status>
    </StackSet>
  </DescribeStackSetResult>
  <ResponseMetadata>
    <RequestId>d8b64e11-5332-46e1-9603-example</RequestId>
  </ResponseMetadata>
</DescribeStackSetResponse>"""

DELETE_STACK_SET_RESPONSE_TEMPLATE = """<DeleteStackSetResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <DeleteStackSetResult/>
  <ResponseMetadata>
    <RequestId>c35ec2d0-d69f-4c4d-9bd7-example</RequestId>
  </ResponseMetadata>
</DeleteStackSetResponse>"""

CREATE_STACK_INSTANCES_TEMPLATE = """<CreateStackInstancesResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <CreateStackInstancesResult>
    <OperationId>1459ad6d-63cc-4c96-a73e-example</OperationId>
  </CreateStackInstancesResult>
  <ResponseMetadata>
    <RequestId>6b29f7e3-69be-4d32-b374-example</RequestId>
  </ResponseMetadata>
</CreateStackInstancesResponse>
"""

LIST_STACK_INSTANCES_TEMPLATE = """<ListStackInstancesResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <ListStackInstancesResult>
    <Summaries>
    {% for instance in stackset.stack_instances %}
      <member>
        <StackId>{{ instance.StackId }}</StackId>
        <StackSetId>{{ instance.StackSetId }}</StackSetId>
        <Region>{{ instance.Region }}</Region>
        <Account>{{ instance.Account }}</Account>
        <Status>{{ instance.Status }}</Status>
      </member>
    {% endfor %}
    </Summaries>
  </ListStackInstancesResult>
  <ResponseMetadata>
    <RequestId>83c27e73-b498-410f-993c-example</RequestId>
  </ResponseMetadata>
</ListStackInstancesResponse>
"""

DELETE_STACK_INSTANCES_TEMPLATE = """<DeleteStackInstancesResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <DeleteStackInstancesResult>
    <OperationId>{{ operation.OperationId }}</OperationId>
  </DeleteStackInstancesResult>
  <ResponseMetadata>
    <RequestId>e5325090-66f6-4ecd-a531-example</RequestId>
  </ResponseMetadata>
</DeleteStackInstancesResponse>
"""

DESCRIBE_STACK_INSTANCE_TEMPLATE = """<DescribeStackInstanceResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <DescribeStackInstanceResult>
    <StackInstance>
      <StackId>{{ instance.StackId }}</StackId>
      <StackSetId>{{ instance.StackSetId }}</StackSetId>
    {% if instance.ParameterOverrides %}
      <ParameterOverrides>
      {% for override in instance.ParameterOverrides %}
      {% if override['ParameterKey'] or override['ParameterValue'] %}
        <member>
          <ParameterKey>{{ override.ParameterKey }}</ParameterKey>
          <UsePreviousValue>false</UsePreviousValue>
          <ParameterValue>{{ override.ParameterValue }}</ParameterValue>
        </member>
      {% endif %}
      {% endfor %}
      </ParameterOverrides>
    {% else %}
      <ParameterOverrides/>
    {% endif %}
      <Region>{{ instance.Region }}</Region>
      <Account>{{ instance.Account }}</Account>
      <Status>{{ instance.Status }}</Status>
    </StackInstance>
  </DescribeStackInstanceResult>
  <ResponseMetadata>
    <RequestId>c6c7be10-0343-4319-8a25-example</RequestId>
  </ResponseMetadata>
</DescribeStackInstanceResponse>
"""

LIST_STACK_SETS_TEMPLATE = """<ListStackSetsResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <ListStackSetsResult>
    <Summaries>
    {% for key, value in stacksets.items() %}
      <member>
        <StackSetName>{{ value.name }}</StackSetName>
        <StackSetId>{{ value.id }}</StackSetId>
        <Status>{{ value.status }}</Status>
      </member>
    {% endfor %}
    </Summaries>
  </ListStackSetsResult>
  <ResponseMetadata>
    <RequestId>4dcacb73-841e-4ed8-b335-example</RequestId>
  </ResponseMetadata>
</ListStackSetsResponse>
"""

UPDATE_STACK_INSTANCES_RESPONSE_TEMPLATE = """<UpdateStackInstancesResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <UpdateStackInstancesResult>
    <OperationId>{{ operation }}</OperationId>
  </UpdateStackInstancesResult>
  <ResponseMetadata>
    <RequestId>bdbf8e94-19b6-4ce4-af85-example</RequestId>
  </ResponseMetadata>
</UpdateStackInstancesResponse>
"""

UPDATE_STACK_SET_RESPONSE_TEMPLATE = """<UpdateStackSetResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <UpdateStackSetResult>
    <OperationId>{{ operation.OperationId }}</OperationId>
  </UpdateStackSetResult>
  <ResponseMetadata>
    <RequestId>adac907b-17e3-43e6-a254-example</RequestId>
  </ResponseMetadata>
</UpdateStackSetResponse>
"""

LIST_STACK_SET_OPERATIONS_RESPONSE_TEMPLATE = """<ListStackSetOperationsResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <ListStackSetOperationsResult>
    <Summaries>
    {% for operation in stackset.operations %}
      <member>
        <CreationTimestamp>{{ operation.CreationTimestamp }}</CreationTimestamp>
        <OperationId>{{ operation.OperationId }}</OperationId>
        <Action>{{ operation.Action }}</Action>
        <EndTimestamp>{{ operation.EndTimestamp }}</EndTimestamp>
        <Status>{{ operation.Status }}</Status>
      </member>
    {% endfor %}
    </Summaries>
  </ListStackSetOperationsResult>
  <ResponseMetadata>
    <RequestId>65b9d9be-08bb-4a43-9a21-example</RequestId>
  </ResponseMetadata>
</ListStackSetOperationsResponse>
"""

STOP_STACK_SET_OPERATION_RESPONSE_TEMPLATE = """<StopStackSetOperationResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <StopStackSetOperationResult/>
  <ResponseMetadata>
    <RequestId>2188554a-07c6-4396-b2c5-example</RequestId>
  </ResponseMetadata>                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     </StopStackSetOperationResponse>
"""

DESCRIBE_STACKSET_OPERATION_RESPONSE_TEMPLATE = """<DescribeStackSetOperationResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <DescribeStackSetOperationResult>
    <StackSetOperation>
      <ExecutionRoleName>{{ stackset.execution_role }}</ExecutionRoleName>
      <AdministrationRoleARN>arn:aws:iam::123456789012:role/{{ stackset.admin_role }}</AdministrationRoleARN>
      <StackSetId>{{ stackset.id }}</StackSetId>
      <CreationTimestamp>{{ operation.CreationTimestamp }}</CreationTimestamp>
      <OperationId>{{ operation.OperationId }}</OperationId>
      <Action>{{ operation.Action }}</Action>
      <OperationPreferences>
        <RegionOrder/>
      </OperationPreferences>
      <EndTimestamp>{{ operation.EndTimestamp }}</EndTimestamp>
      <Status>{{ operation.Status }}</Status>
    </StackSetOperation>
  </DescribeStackSetOperationResult>
  <ResponseMetadata>
    <RequestId>2edc27b6-9ce2-486a-a192-example</RequestId>
  </ResponseMetadata>
</DescribeStackSetOperationResponse>
"""

LIST_STACK_SET_OPERATION_RESULTS_RESPONSE_TEMPLATE = """<ListStackSetOperationResultsResponse xmlns="http://internal.amazon.com/coral/com.amazonaws.maestro.service.v20160713/">
  <ListStackSetOperationResultsResult>
    <Summaries>
    {% for instance in operation.Instances %}
    {% for account, region in instance.items() %}
      <member>
        <AccountGateResult>
          <StatusReason>Function not found: arn:aws:lambda:us-west-2:123456789012:function:AWSCloudFormationStackSetAccountGate</StatusReason>
          <Status>SKIPPED</Status>
        </AccountGateResult>
        <Region>{{ region }}</Region>
        <Account>{{ account }}</Account>
        <Status>{{ operation.Status }}</Status>
      </member>
    {% endfor %}
    {% endfor %}
    </Summaries>
  </ListStackSetOperationResultsResult>
  <ResponseMetadata>
    <RequestId>ac05a9ce-5f98-4197-a29b-example</RequestId>
  </ResponseMetadata>
</ListStackSetOperationResultsResponse>
"""



from __future__ import unicode_literals
import os
from sure import expect
from moto.s3.utils import bucket_name_from_url, _VersionedKeyStore, parse_region_from_url


def test_base_url():
    expect(bucket_name_from_url('https://s3.amazonaws.com/')).should.equal(None)


def test_localhost_bucket():
    expect(bucket_name_from_url('https://wfoobar.localhost:5000/abc')
           ).should.equal("wfoobar")


def test_localhost_without_bucket():
    expect(bucket_name_from_url(
        'https://www.localhost:5000/def')).should.equal(None)

def test_force_ignore_subdomain_for_bucketnames():
    os.environ['S3_IGNORE_SUBDOMAIN_BUCKETNAME'] = '1'
    expect(bucket_name_from_url('https://subdomain.localhost:5000/abc/resource')).should.equal(None)
    del(os.environ['S3_IGNORE_SUBDOMAIN_BUCKETNAME'])



def test_versioned_key_store():
    d = _VersionedKeyStore()

    d.should.have.length_of(0)

    d['key'] = [1]

    d.should.have.length_of(1)

    d['key'] = 2
    d.should.have.length_of(1)

    d.should.have.key('key').being.equal(2)

    d.get.when.called_with('key').should.return_value(2)
    d.get.when.called_with('badkey').should.return_value(None)
    d.get.when.called_with('badkey', 'HELLO').should.return_value('HELLO')

    # Tests key[
    d.shouldnt.have.key('badkey')
    d.__getitem__.when.called_with('badkey').should.throw(KeyError)

    d.getlist('key').should.have.length_of(2)
    d.getlist('key').should.be.equal([[1], 2])
    d.getlist('badkey').should.be.none

    d.setlist('key', 1)
    d.getlist('key').should.be.equal([1])

    d.setlist('key', (1, 2))
    d.getlist('key').shouldnt.be.equal((1, 2))
    d.getlist('key').should.be.equal([1, 2])

    d.setlist('key', [[1], [2]])
    d['key'].should.have.length_of(1)
    d.getlist('key').should.be.equal([[1], [2]])


def test_parse_region_from_url():
    expected = 'us-west-2'
    for url in ['http://s3-us-west-2.amazonaws.com/bucket',
                'http://s3.us-west-2.amazonaws.com/bucket',
                'http://bucket.s3-us-west-2.amazonaws.com',
                'https://s3-us-west-2.amazonaws.com/bucket',
                'https://s3.us-west-2.amazonaws.com/bucket',
                'https://bucket.s3-us-west-2.amazonaws.com']:
        parse_region_from_url(url).should.equal(expected)

    expected = 'us-east-1'
    for url in ['http://s3.amazonaws.com/bucket',
                'http://bucket.s3.amazonaws.com',
                'https://s3.amazonaws.com/bucket',
                'https://bucket.s3.amazonaws.com']:
        parse_region_from_url(url).should.equal(expected)



from __future__ import unicode_literals

import time
import datetime
import boto3
from botocore.exceptions import ClientError
import sure  # noqa
from moto import mock_batch, mock_iam, mock_ec2, mock_ecs, mock_logs, mock_cloudformation
import functools
import nose
import json

DEFAULT_REGION = 'eu-central-1'


def _get_clients():
    return boto3.client('ec2', region_name=DEFAULT_REGION), \
           boto3.client('iam', region_name=DEFAULT_REGION), \
           boto3.client('ecs', region_name=DEFAULT_REGION), \
           boto3.client('logs', region_name=DEFAULT_REGION), \
           boto3.client('batch', region_name=DEFAULT_REGION)


def _setup(ec2_client, iam_client):
    """
    Do prerequisite setup
    :return: VPC ID, Subnet ID, Security group ID, IAM Role ARN
    :rtype: tuple
    """
    resp = ec2_client.create_vpc(CidrBlock='172.30.0.0/24')
    vpc_id = resp['Vpc']['VpcId']
    resp = ec2_client.create_subnet(
        AvailabilityZone='eu-central-1a',
        CidrBlock='172.30.0.0/25',
        VpcId=vpc_id
    )
    subnet_id = resp['Subnet']['SubnetId']
    resp = ec2_client.create_security_group(
        Description='test_sg_desc',
        GroupName='test_sg',
        VpcId=vpc_id
    )
    sg_id = resp['GroupId']

    resp = iam_client.create_role(
        RoleName='TestRole',
        AssumeRolePolicyDocument='some_policy'
    )
    iam_arn = resp['Role']['Arn']

    return vpc_id, subnet_id, sg_id, iam_arn


@mock_cloudformation()
@mock_ec2
@mock_ecs
@mock_iam
@mock_batch
def test_create_env_cf():
    ec2_client, iam_client, ecs_client, logs_client, batch_client = _get_clients()
    vpc_id, subnet_id, sg_id, iam_arn = _setup(ec2_client, iam_client)

    create_environment_template = {
        'Resources': {
            "ComputeEnvironment": {
                "Type": "AWS::Batch::ComputeEnvironment",
                "Properties": {
                    "Type": "MANAGED",
                    "ComputeResources": {
                        "Type": "EC2",
                        "MinvCpus": 0,
                        "DesiredvCpus": 0,
                        "MaxvCpus": 64,
                        "InstanceTypes": [
                            "optimal"
                        ],
                        "Subnets": [subnet_id],
                        "SecurityGroupIds": [sg_id],
                        "InstanceRole": iam_arn
                    },
                    "ServiceRole": iam_arn
                }
            }
        }
    }
    cf_json = json.dumps(create_environment_template)

    cf_conn = boto3.client('cloudformation', DEFAULT_REGION)
    stack_id = cf_conn.create_stack(
        StackName='test_stack',
        TemplateBody=cf_json,
    )['StackId']

    stack_resources = cf_conn.list_stack_resources(StackName=stack_id)

    stack_resources['StackResourceSummaries'][0]['ResourceStatus'].should.equal('CREATE_COMPLETE')
    # Spot checks on the ARN
    stack_resources['StackResourceSummaries'][0]['PhysicalResourceId'].startswith('arn:aws:batch:')
    stack_resources['StackResourceSummaries'][0]['PhysicalResourceId'].should.contain('test_stack')


@mock_cloudformation()
@mock_ec2
@mock_ecs
@mock_iam
@mock_batch
def test_create_job_queue_cf():
    ec2_client, iam_client, ecs_client, logs_client, batch_client = _get_clients()
    vpc_id, subnet_id, sg_id, iam_arn = _setup(ec2_client, iam_client)

    create_environment_template = {
        'Resources': {
            "ComputeEnvironment": {
                "Type": "AWS::Batch::ComputeEnvironment",
                "Properties": {
                    "Type": "MANAGED",
                    "ComputeResources": {
                        "Type": "EC2",
                        "MinvCpus": 0,
                        "DesiredvCpus": 0,
                        "MaxvCpus": 64,
                        "InstanceTypes": [
                            "optimal"
                        ],
                        "Subnets": [subnet_id],
                        "SecurityGroupIds": [sg_id],
                        "InstanceRole": iam_arn
                    },
                    "ServiceRole": iam_arn
                }
            },

            "JobQueue": {
                "Type": "AWS::Batch::JobQueue",
                "Properties": {
                    "Priority": 1,
                    "ComputeEnvironmentOrder": [
                        {
                            "Order": 1,
                            "ComputeEnvironment": {"Ref": "ComputeEnvironment"}
                        }
                    ]
                }
            },
        }
    }
    cf_json = json.dumps(create_environment_template)

    cf_conn = boto3.client('cloudformation', DEFAULT_REGION)
    stack_id = cf_conn.create_stack(
        StackName='test_stack',
        TemplateBody=cf_json,
    )['StackId']

    stack_resources = cf_conn.list_stack_resources(StackName=stack_id)
    len(stack_resources['StackResourceSummaries']).should.equal(2)

    job_queue_resource = list(filter(lambda item: item['ResourceType'] == 'AWS::Batch::JobQueue', stack_resources['StackResourceSummaries']))[0]

    job_queue_resource['ResourceStatus'].should.equal('CREATE_COMPLETE')
    # Spot checks on the ARN
    job_queue_resource['PhysicalResourceId'].startswith('arn:aws:batch:')
    job_queue_resource['PhysicalResourceId'].should.contain('test_stack')
    job_queue_resource['PhysicalResourceId'].should.contain('job-queue/')


@mock_cloudformation()
@mock_ec2
@mock_ecs
@mock_iam
@mock_batch
def test_create_job_def_cf():
    ec2_client, iam_client, ecs_client, logs_client, batch_client = _get_clients()
    vpc_id, subnet_id, sg_id, iam_arn = _setup(ec2_client, iam_client)

    create_environment_template = {
        'Resources': {
            "ComputeEnvironment": {
                "Type": "AWS::Batch::ComputeEnvironment",
                "Properties": {
                    "Type": "MANAGED",
                    "ComputeResources": {
                        "Type": "EC2",
                        "MinvCpus": 0,
                        "DesiredvCpus": 0,
                        "MaxvCpus": 64,
                        "InstanceTypes": [
                            "optimal"
                        ],
                        "Subnets": [subnet_id],
                        "SecurityGroupIds": [sg_id],
                        "InstanceRole": iam_arn
                    },
                    "ServiceRole": iam_arn
                }
            },

            "JobQueue": {
                "Type": "AWS::Batch::JobQueue",
                "Properties": {
                    "Priority": 1,
                    "ComputeEnvironmentOrder": [
                        {
                            "Order": 1,
                            "ComputeEnvironment": {"Ref": "ComputeEnvironment"}
                        }
                    ]
                }
            },

            "JobDefinition": {
                "Type": "AWS::Batch::JobDefinition",
                "Properties": {
                    "Type": "container",
                    "ContainerProperties": {
                        "Image": {
                            "Fn::Join": ["", ["137112412989.dkr.ecr.", {"Ref": "AWS::Region"}, ".amazonaws.com/amazonlinux:latest"]]
                        },
                        "Vcpus": 2,
                        "Memory": 2000,
                        "Command": ["echo", "Hello world"]
                    },
                    "RetryStrategy": {
                        "Attempts": 1
                    }
                }
            },
        }
    }
    cf_json = json.dumps(create_environment_template)

    cf_conn = boto3.client('cloudformation', DEFAULT_REGION)
    stack_id = cf_conn.create_stack(
        StackName='test_stack',
        TemplateBody=cf_json,
    )['StackId']

    stack_resources = cf_conn.list_stack_resources(StackName=stack_id)
    len(stack_resources['StackResourceSummaries']).should.equal(3)

    job_def_resource = list(filter(lambda item: item['ResourceType'] == 'AWS::Batch::JobDefinition', stack_resources['StackResourceSummaries']))[0]

    job_def_resource['ResourceStatus'].should.equal('CREATE_COMPLETE')
    # Spot checks on the ARN
    job_def_resource['PhysicalResourceId'].startswith('arn:aws:batch:')
    job_def_resource['PhysicalResourceId'].should.contain('test_stack-JobDef')
    job_def_resource['PhysicalResourceId'].should.contain('job-definition/')



from __future__ import unicode_literals
import json
from base64 import b64encode
from datetime import datetime
import time

from moto.core.responses import BaseResponse
from .models import ecr_backends, DEFAULT_REGISTRY_ID


class ECRResponse(BaseResponse):
    @property
    def ecr_backend(self):
        return ecr_backends[self.region]

    @property
    def request_params(self):
        try:
            return json.loads(self.body)
        except ValueError:
            return {}

    def _get_param(self, param):
        return self.request_params.get(param, None)

    def create_repository(self):
        repository_name = self._get_param('repositoryName')
        if repository_name is None:
            repository_name = 'default'
        repository = self.ecr_backend.create_repository(repository_name)
        return json.dumps({
            'repository': repository.response_object
        })

    def describe_repositories(self):
        describe_repositories_name = self._get_param('repositoryNames')
        registry_id = self._get_param('registryId')

        repositories = self.ecr_backend.describe_repositories(
            repository_names=describe_repositories_name, registry_id=registry_id)
        return json.dumps({
            'repositories': repositories,
            'failures': []
        })

    def delete_repository(self):
        repository_str = self._get_param('repositoryName')
        registry_id = self._get_param('registryId')
        repository = self.ecr_backend.delete_repository(repository_str, registry_id)
        return json.dumps({
            'repository': repository.response_object
        })

    def put_image(self):
        repository_str = self._get_param('repositoryName')
        image_manifest = self._get_param('imageManifest')
        image_tag = self._get_param('imageTag')
        image = self.ecr_backend.put_image(repository_str, image_manifest, image_tag)

        return json.dumps({
            'image': image.response_object
        })

    def list_images(self):
        repository_str = self._get_param('repositoryName')
        registry_id = self._get_param('registryId')
        images = self.ecr_backend.list_images(repository_str, registry_id)
        return json.dumps({
            'imageIds': [image.response_list_object for image in images],
        })

    def describe_images(self):
        repository_str = self._get_param('repositoryName')
        registry_id = self._get_param('registryId')
        image_ids = self._get_param('imageIds')
        images = self.ecr_backend.describe_images(repository_str, registry_id, image_ids)
        return json.dumps({
            'imageDetails': [image.response_describe_object for image in images],
        })

    def batch_check_layer_availability(self):
        if self.is_not_dryrun('BatchCheckLayerAvailability'):
            raise NotImplementedError(
                'ECR.batch_check_layer_availability is not yet implemented')

    def batch_delete_image(self):
        repository_str = self._get_param('repositoryName')
        registry_id = self._get_param('registryId')
        image_ids = self._get_param('imageIds')

        response = self.ecr_backend.batch_delete_image(repository_str, registry_id, image_ids)
        return json.dumps(response)

    def batch_get_image(self):
        repository_str = self._get_param('repositoryName')
        registry_id = self._get_param('registryId')
        image_ids = self._get_param('imageIds')
        accepted_media_types = self._get_param('acceptedMediaTypes')

        response = self.ecr_backend.batch_get_image(repository_str, registry_id, image_ids, accepted_media_types)
        return json.dumps(response)

    def can_paginate(self):
        if self.is_not_dryrun('CanPaginate'):
            raise NotImplementedError(
                'ECR.can_paginate is not yet implemented')

    def complete_layer_upload(self):
        if self.is_not_dryrun('CompleteLayerUpload'):
            raise NotImplementedError(
                'ECR.complete_layer_upload is not yet implemented')

    def delete_repository_policy(self):
        if self.is_not_dryrun('DeleteRepositoryPolicy'):
            raise NotImplementedError(
                'ECR.delete_repository_policy is not yet implemented')

    def generate_presigned_url(self):
        if self.is_not_dryrun('GeneratePresignedUrl'):
            raise NotImplementedError(
                'ECR.generate_presigned_url is not yet implemented')

    def get_authorization_token(self):
        registry_ids = self._get_param('registryIds')
        if not registry_ids:
            registry_ids = [DEFAULT_REGISTRY_ID]
        auth_data = []
        for registry_id in registry_ids:
            password = '{}-auth-token'.format(registry_id)
            auth_token = b64encode("AWS:{}".format(password).encode('ascii')).decode()
            auth_data.append({
                'authorizationToken': auth_token,
                'expiresAt': time.mktime(datetime(2015, 1, 1).timetuple()),
                'proxyEndpoint': 'https://{}.dkr.ecr.{}.amazonaws.com'.format(registry_id, self.region)
            })
        return json.dumps({'authorizationData': auth_data})

    def get_download_url_for_layer(self):
        if self.is_not_dryrun('GetDownloadUrlForLayer'):
            raise NotImplementedError(
                'ECR.get_download_url_for_layer is not yet implemented')

    def get_paginator(self):
        if self.is_not_dryrun('GetPaginator'):
            raise NotImplementedError(
                'ECR.get_paginator is not yet implemented')

    def get_repository_policy(self):
        if self.is_not_dryrun('GetRepositoryPolicy'):
            raise NotImplementedError(
                'ECR.get_repository_policy is not yet implemented')

    def get_waiter(self):
        if self.is_not_dryrun('GetWaiter'):
            raise NotImplementedError(
                'ECR.get_waiter is not yet implemented')

    def initiate_layer_upload(self):
        if self.is_not_dryrun('InitiateLayerUpload'):
            raise NotImplementedError(
                'ECR.initiate_layer_upload is not yet implemented')

    def set_repository_policy(self):
        if self.is_not_dryrun('SetRepositoryPolicy'):
            raise NotImplementedError(
                'ECR.set_repository_policy is not yet implemented')

    def upload_layer_part(self):
        if self.is_not_dryrun('UploadLayerPart'):
            raise NotImplementedError(
                'ECR.upload_layer_part is not yet implemented')



from __future__ import unicode_literals

import os
import json

import boto
import boto.s3
import boto.s3.key
import boto.cloudformation
from boto.exception import BotoServerError
import sure  # noqa
# Ensure 'assert_raises' context manager support for Python 2.6
import tests.backport_assert_raises  # noqa
from nose.tools import assert_raises

from moto import mock_cloudformation_deprecated, mock_s3_deprecated, mock_route53_deprecated
from moto.cloudformation import cloudformation_backends

dummy_template = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Description": "Stack 1",
    "Resources": {},
}

dummy_template2 = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Description": "Stack 2",
    "Resources": {},
}

# template with resource which has no delete attribute defined
dummy_template3 = {
    "AWSTemplateFormatVersion": "2010-09-09",
    "Description": "Stack 3",
    "Resources": {
        "VPC": {
            "Properties": {
                "CidrBlock": "192.168.0.0/16",
            },
            "Type": "AWS::EC2::VPC"
        }
    },
}

dummy_template_json = json.dumps(dummy_template)
dummy_template_json2 = json.dumps(dummy_template2)
dummy_template_json3 = json.dumps(dummy_template3)


@mock_cloudformation_deprecated
def test_create_stack():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    stack = conn.describe_stacks()[0]
    stack.stack_name.should.equal('test_stack')
    stack.get_template().should.equal({
        'GetTemplateResponse': {
            'GetTemplateResult': {
                'TemplateBody': dummy_template_json,
                'ResponseMetadata': {
                    'RequestId': '2d06e36c-ac1d-11e0-a958-f9382b6eb86bEXAMPLE'
                }
            }
        }

    })


@mock_cloudformation_deprecated
@mock_route53_deprecated
def test_create_stack_hosted_zone_by_id():
    conn = boto.connect_cloudformation()
    dummy_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Stack 1",
        "Parameters": {
        },
        "Resources": {
            "Bar": {
                "Type" : "AWS::Route53::HostedZone",
                "Properties" : {
                    "Name" : "foo.bar.baz",
                }
            },
        },
    }
    dummy_template2 = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Stack 2",
        "Parameters": {
            "ZoneId": { "Type": "String" }
        },
        "Resources": {
            "Foo": {
                "Properties": {
                    "HostedZoneId": {"Ref": "ZoneId"},
                    "RecordSets": []
                },
                "Type": "AWS::Route53::RecordSetGroup"
            }
        },
    }
    conn.create_stack(
        "test_stack",
        template_body=json.dumps(dummy_template),
        parameters={}.items()
    )
    r53_conn = boto.connect_route53()
    zone_id = r53_conn.get_zones()[0].id
    conn.create_stack(
        "test_stack",
        template_body=json.dumps(dummy_template2),
        parameters={"ZoneId": zone_id}.items()
    )

    stack = conn.describe_stacks()[0]
    assert stack.list_resources()


@mock_cloudformation_deprecated
def test_creating_stacks_across_regions():
    west1_conn = boto.cloudformation.connect_to_region("us-west-1")
    west1_conn.create_stack("test_stack", template_body=dummy_template_json)

    west2_conn = boto.cloudformation.connect_to_region("us-west-2")
    west2_conn.create_stack("test_stack", template_body=dummy_template_json)

    list(west1_conn.describe_stacks()).should.have.length_of(1)
    list(west2_conn.describe_stacks()).should.have.length_of(1)


@mock_cloudformation_deprecated
def test_create_stack_with_notification_arn():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack_with_notifications",
        template_body=dummy_template_json,
        notification_arns='arn:aws:sns:us-east-1:123456789012:fake-queue'
    )

    stack = conn.describe_stacks()[0]
    [n.value for n in stack.notification_arns].should.contain(
        'arn:aws:sns:us-east-1:123456789012:fake-queue')


@mock_cloudformation_deprecated
@mock_s3_deprecated
def test_create_stack_from_s3_url():
    s3_conn = boto.s3.connect_to_region('us-west-1')
    bucket = s3_conn.create_bucket("foobar")
    key = boto.s3.key.Key(bucket)
    key.key = "template-key"
    key.set_contents_from_string(dummy_template_json)
    key_url = key.generate_url(expires_in=0, query_auth=False)

    conn = boto.cloudformation.connect_to_region('us-west-1')
    conn.create_stack('new-stack', template_url=key_url)

    stack = conn.describe_stacks()[0]
    stack.stack_name.should.equal('new-stack')
    stack.get_template().should.equal(
        {
            'GetTemplateResponse': {
                'GetTemplateResult': {
                    'TemplateBody': dummy_template_json,
                    'ResponseMetadata': {
                        'RequestId': '2d06e36c-ac1d-11e0-a958-f9382b6eb86bEXAMPLE'
                    }
                }
            }

        })


@mock_cloudformation_deprecated
def test_describe_stack_by_name():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    stack = conn.describe_stacks("test_stack")[0]
    stack.stack_name.should.equal('test_stack')


@mock_cloudformation_deprecated
def test_describe_stack_by_stack_id():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    stack = conn.describe_stacks("test_stack")[0]
    stack_by_id = conn.describe_stacks(stack.stack_id)[0]
    stack_by_id.stack_id.should.equal(stack.stack_id)
    stack_by_id.stack_name.should.equal("test_stack")


@mock_cloudformation_deprecated
def test_describe_deleted_stack():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    stack = conn.describe_stacks("test_stack")[0]
    stack_id = stack.stack_id
    conn.delete_stack(stack.stack_id)
    stack_by_id = conn.describe_stacks(stack_id)[0]
    stack_by_id.stack_id.should.equal(stack.stack_id)
    stack_by_id.stack_name.should.equal("test_stack")
    stack_by_id.stack_status.should.equal("DELETE_COMPLETE")


@mock_cloudformation_deprecated
def test_get_template_by_name():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    template = conn.get_template("test_stack")
    template.should.equal({
        'GetTemplateResponse': {
            'GetTemplateResult': {
                'TemplateBody': dummy_template_json,
                'ResponseMetadata': {
                    'RequestId': '2d06e36c-ac1d-11e0-a958-f9382b6eb86bEXAMPLE'
                }
            }
        }

    })


@mock_cloudformation_deprecated
def test_list_stacks():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )
    conn.create_stack(
        "test_stack2",
        template_body=dummy_template_json,
    )

    stacks = conn.list_stacks()
    stacks.should.have.length_of(2)
    stacks[0].template_description.should.equal("Stack 1")


@mock_cloudformation_deprecated
def test_delete_stack_by_name():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    conn.describe_stacks().should.have.length_of(1)
    conn.delete_stack("test_stack")
    conn.describe_stacks().should.have.length_of(0)


@mock_cloudformation_deprecated
def test_delete_stack_by_id():
    conn = boto.connect_cloudformation()
    stack_id = conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    conn.describe_stacks().should.have.length_of(1)
    conn.delete_stack(stack_id)
    conn.describe_stacks().should.have.length_of(0)
    with assert_raises(BotoServerError):
        conn.describe_stacks("test_stack")

    conn.describe_stacks(stack_id).should.have.length_of(1)


@mock_cloudformation_deprecated
def test_delete_stack_with_resource_missing_delete_attr():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json3,
    )

    conn.describe_stacks().should.have.length_of(1)
    conn.delete_stack("test_stack")
    conn.describe_stacks().should.have.length_of(0)


@mock_cloudformation_deprecated
def test_bad_describe_stack():
    conn = boto.connect_cloudformation()
    with assert_raises(BotoServerError):
        conn.describe_stacks("bad_stack")


@mock_cloudformation_deprecated()
def test_cloudformation_params():
    dummy_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Stack 1",
        "Resources": {},
        "Parameters": {
            "APPNAME": {
                "Default": "app-name",
                "Description": "The name of the app",
                "Type": "String"
            }
        }
    }
    dummy_template_json = json.dumps(dummy_template)
    cfn = boto.connect_cloudformation()
    cfn.create_stack('test_stack1', template_body=dummy_template_json, parameters=[
                     ('APPNAME', 'testing123')])
    stack = cfn.describe_stacks('test_stack1')[0]
    stack.parameters.should.have.length_of(1)
    param = stack.parameters[0]
    param.key.should.equal('APPNAME')
    param.value.should.equal('testing123')


@mock_cloudformation_deprecated
def test_cloudformation_params_conditions_and_resources_are_distinct():
    dummy_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Stack 1",
        "Conditions": {
            "FooEnabled": {
                "Fn::Equals": [
                    {
                        "Ref": "FooEnabled"
                    },
                    "true"
                ]
            },
            "FooDisabled": {
                "Fn::Not": [
                    {
                        "Fn::Equals": [
                            {
                                "Ref": "FooEnabled"
                            },
                            "true"
                        ]
                    }
                ]
            }
        },
        "Parameters": {
            "FooEnabled": {
                "Type": "String",
                "AllowedValues": [
                    "true",
                    "false"
                ]
            }
        },
        "Resources": {
            "Bar": {
                "Properties": {
                    "CidrBlock": "192.168.0.0/16",
                },
                "Condition": "FooDisabled",
                "Type": "AWS::EC2::VPC"
            }
        }
    }
    dummy_template_json = json.dumps(dummy_template)
    cfn = boto.connect_cloudformation()
    cfn.create_stack('test_stack1', template_body=dummy_template_json, parameters=[('FooEnabled', 'true')])
    stack = cfn.describe_stacks('test_stack1')[0]
    resources = stack.list_resources()
    assert not [resource for resource in resources if resource.logical_resource_id == 'Bar']


@mock_cloudformation_deprecated
def test_stack_tags():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
        tags={"foo": "bar", "baz": "bleh"},
    )

    stack = conn.describe_stacks()[0]
    dict(stack.tags).should.equal({"foo": "bar", "baz": "bleh"})


@mock_cloudformation_deprecated
def test_update_stack():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )

    conn.update_stack("test_stack", dummy_template_json2)

    stack = conn.describe_stacks()[0]
    stack.stack_status.should.equal("UPDATE_COMPLETE")
    stack.get_template().should.equal({
        'GetTemplateResponse': {
            'GetTemplateResult': {
                'TemplateBody': dummy_template_json2,
                'ResponseMetadata': {
                    'RequestId': '2d06e36c-ac1d-11e0-a958-f9382b6eb86bEXAMPLE'
                }
            }
        }
    })


@mock_cloudformation_deprecated
def test_update_stack_with_previous_template():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
    )
    conn.update_stack("test_stack", use_previous_template=True)

    stack = conn.describe_stacks()[0]
    stack.stack_status.should.equal("UPDATE_COMPLETE")
    stack.get_template().should.equal({
        'GetTemplateResponse': {
            'GetTemplateResult': {
                'TemplateBody': dummy_template_json,
                'ResponseMetadata': {
                    'RequestId': '2d06e36c-ac1d-11e0-a958-f9382b6eb86bEXAMPLE'
                }
            }
        }
    })


@mock_cloudformation_deprecated
def test_update_stack_with_parameters():
    dummy_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Stack",
        "Resources": {
            "VPC": {
                "Properties": {
                    "CidrBlock": {"Ref": "Bar"}
            },
            "Type": "AWS::EC2::VPC"
            }
        },
        "Parameters": {
            "Bar": {
                "Type": "String"
            }
        }
    }
    dummy_template_json = json.dumps(dummy_template)
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
        parameters=[("Bar", "192.168.0.0/16")]
    )
    conn.update_stack(
        "test_stack",
        template_body=dummy_template_json,
        parameters=[("Bar", "192.168.0.1/16")]
    )

    stack = conn.describe_stacks()[0]
    assert stack.parameters[0].value == "192.168.0.1/16"


@mock_cloudformation_deprecated
def test_update_stack_replace_tags():
    conn = boto.connect_cloudformation()
    conn.create_stack(
        "test_stack",
        template_body=dummy_template_json,
        tags={"foo": "bar"},
    )
    conn.update_stack(
        "test_stack",
        template_body=dummy_template_json,
        tags={"foo": "baz"},
    )

    stack = conn.describe_stacks()[0]
    stack.stack_status.should.equal("UPDATE_COMPLETE")
    # since there is one tag it doesn't come out as a list
    dict(stack.tags).should.equal({"foo": "baz"})


@mock_cloudformation_deprecated
def test_update_stack_when_rolled_back():
    conn = boto.connect_cloudformation()
    stack_id = conn.create_stack(
        "test_stack", template_body=dummy_template_json)

    cloudformation_backends[conn.region.name].stacks[
        stack_id].status = 'ROLLBACK_COMPLETE'

    with assert_raises(BotoServerError) as err:
        conn.update_stack("test_stack", dummy_template_json)

    ex = err.exception
    ex.body.should.match(
        r'is in ROLLBACK_COMPLETE state and can not be updated')
    ex.error_code.should.equal('ValidationError')
    ex.reason.should.equal('Bad Request')
    ex.status.should.equal(400)


@mock_cloudformation_deprecated
def test_describe_stack_events_shows_create_update_and_delete():
    conn = boto.connect_cloudformation()
    stack_id = conn.create_stack(
        "test_stack", template_body=dummy_template_json)
    conn.update_stack(stack_id, template_body=dummy_template_json2)
    conn.delete_stack(stack_id)

    # assert begins and ends with stack events
    events = conn.describe_stack_events(stack_id)
    events[0].resource_type.should.equal("AWS::CloudFormation::Stack")
    events[-1].resource_type.should.equal("AWS::CloudFormation::Stack")

    # testing ordering of stack events without assuming resource events will not exist
    # the AWS API returns events in reverse chronological order
    stack_events_to_look_for = iter([
        ("DELETE_COMPLETE", None),
        ("DELETE_IN_PROGRESS", "User Initiated"),
        ("UPDATE_COMPLETE", None),
        ("UPDATE_IN_PROGRESS", "User Initiated"),
        ("CREATE_COMPLETE", None),
        ("CREATE_IN_PROGRESS", "User Initiated"),
    ])
    try:
        for event in events:
            event.stack_id.should.equal(stack_id)
            event.stack_name.should.equal("test_stack")
            event.event_id.should.match(r"[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}")

            if event.resource_type == "AWS::CloudFormation::Stack":
                event.logical_resource_id.should.equal("test_stack")
                event.physical_resource_id.should.equal(stack_id)

                status_to_look_for, reason_to_look_for = next(
                    stack_events_to_look_for)
                event.resource_status.should.equal(status_to_look_for)
                if reason_to_look_for is not None:
                    event.resource_status_reason.should.equal(
                        reason_to_look_for)
    except StopIteration:
        assert False, "Too many stack events"

    list(stack_events_to_look_for).should.be.empty


@mock_cloudformation_deprecated
def test_create_stack_lambda_and_dynamodb():
    conn = boto.connect_cloudformation()
    dummy_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Stack Lambda Test 1",
        "Parameters": {
        },
        "Resources": {
            "func1": {
                "Type" : "AWS::Lambda::Function",
                "Properties" : {
                    "Code": {
                        "S3Bucket": "bucket_123",
                        "S3Key": "key_123"
                    },
                    "FunctionName": "func1",
                    "Handler": "handler.handler",
                    "Role": "role1",
                    "Runtime": "python2.7",
                    "Description": "descr",
                    "MemorySize": 12345,
                }
            },
            "func1version": {
                "Type": "AWS::Lambda::LambdaVersion",
                "Properties" : {
                    "Version": "v1.2.3"
                }
            },
            "tab1": {
                "Type" : "AWS::DynamoDB::Table",
                "Properties" : {
                    "TableName": "tab1",
                    "KeySchema": [{
                        "AttributeName": "attr1",
                        "KeyType": "HASH"
                    }],
                    "AttributeDefinitions": [{
                        "AttributeName": "attr1",
                        "AttributeType": "string"
                    }],
                    "ProvisionedThroughput": {
                        "ReadCapacityUnits": 10,
                        "WriteCapacityUnits": 10
                    }
                }
            },
            "func1mapping": {
                "Type": "AWS::Lambda::EventSourceMapping",
                "Properties" : {
                    "FunctionName": "v1.2.3",
                    "EventSourceArn": "arn:aws:dynamodb:region:XXXXXX:table/tab1/stream/2000T00:00:00.000",
                    "StartingPosition": "0",
                    "BatchSize": 100,
                    "Enabled": True
                }
            }
        },
    }
    validate_s3_before = os.environ.get('VALIDATE_LAMBDA_S3', '')
    try:
        os.environ['VALIDATE_LAMBDA_S3'] = 'false'
        conn.create_stack(
            "test_stack_lambda_1",
            template_body=json.dumps(dummy_template),
            parameters={}.items()
        )
    finally:
        os.environ['VALIDATE_LAMBDA_S3'] = validate_s3_before

    stack = conn.describe_stacks()[0]
    resources = stack.list_resources()
    assert len(resources) == 4


@mock_cloudformation_deprecated
def test_create_stack_kinesis():
    conn = boto.connect_cloudformation()
    dummy_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "Stack Kinesis Test 1",
        "Parameters": {},
        "Resources": {
            "stream1": {
                "Type" : "AWS::Kinesis::Stream",
                "Properties" : {
                    "Name": "stream1",
                    "ShardCount": 2
                }
            }
        }
    }
    conn.create_stack(
        "test_stack_kinesis_1",
        template_body=json.dumps(dummy_template),
        parameters={}.items()
    )

    stack = conn.describe_stacks()[0]
    resources = stack.list_resources()
    assert len(resources) == 1



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import time
from copy import deepcopy
from datetime import datetime

import boto3
import pytz
import six
import sure  # noqa
from botocore.exceptions import ClientError
from nose.tools import assert_raises

from moto import mock_emr


run_job_flow_args = dict(
    Instances={
        'InstanceCount': 3,
        'KeepJobFlowAliveWhenNoSteps': True,
        'MasterInstanceType': 'c3.medium',
        'Placement': {'AvailabilityZone': 'us-east-1a'},
        'SlaveInstanceType': 'c3.xlarge',
    },
    JobFlowRole='EMR_EC2_DefaultRole',
    LogUri='s3://mybucket/log',
    Name='cluster',
    ServiceRole='EMR_DefaultRole',
    VisibleToAllUsers=True)


input_instance_groups = [
    {'InstanceCount': 1,
     'InstanceRole': 'MASTER',
     'InstanceType': 'c1.medium',
     'Market': 'ON_DEMAND',
     'Name': 'master'},
    {'InstanceCount': 3,
     'InstanceRole': 'CORE',
     'InstanceType': 'c1.medium',
     'Market': 'ON_DEMAND',
     'Name': 'core'},
    {'InstanceCount': 6,
     'InstanceRole': 'TASK',
     'InstanceType': 'c1.large',
     'Market': 'SPOT',
     'Name': 'task-1',
     'BidPrice': '0.07'},
    {'InstanceCount': 10,
     'InstanceRole': 'TASK',
     'InstanceType': 'c1.xlarge',
     'Market': 'SPOT',
     'Name': 'task-2',
     'BidPrice': '0.05'},
]


@mock_emr
def test_describe_cluster():
    client = boto3.client('emr', region_name='us-east-1')

    args = deepcopy(run_job_flow_args)
    args['Applications'] = [{'Name': 'Spark', 'Version': '2.4.2'}]
    args['Configurations'] = [
        {'Classification': 'yarn-site',
         'Properties': {'someproperty': 'somevalue',
                        'someotherproperty': 'someothervalue'}},
        {'Classification': 'nested-configs',
         'Properties': {},
         'Configurations': [
             {
                 'Classification': 'nested-config',
                 'Properties': {
                     'nested-property': 'nested-value'
                 }
             }
         ]}
    ]
    args['Instances']['AdditionalMasterSecurityGroups'] = ['additional-master']
    args['Instances']['AdditionalSlaveSecurityGroups'] = ['additional-slave']
    args['Instances']['Ec2KeyName'] = 'mykey'
    args['Instances']['Ec2SubnetId'] = 'subnet-8be41cec'
    args['Instances']['EmrManagedMasterSecurityGroup'] = 'master-security-group'
    args['Instances']['EmrManagedSlaveSecurityGroup'] = 'slave-security-group'
    args['Instances']['KeepJobFlowAliveWhenNoSteps'] = False
    args['Instances']['ServiceAccessSecurityGroup'] = 'service-access-security-group'
    args['Tags'] = [{'Key': 'tag1', 'Value': 'val1'},
                    {'Key': 'tag2', 'Value': 'val2'}]

    cluster_id = client.run_job_flow(**args)['JobFlowId']

    cl = client.describe_cluster(ClusterId=cluster_id)['Cluster']
    cl['Applications'][0]['Name'].should.equal('Spark')
    cl['Applications'][0]['Version'].should.equal('2.4.2')
    cl['AutoTerminate'].should.equal(True)

    config = cl['Configurations'][0]
    config['Classification'].should.equal('yarn-site')
    config['Properties'].should.equal(args['Configurations'][0]['Properties'])

    nested_config = cl['Configurations'][1]
    nested_config['Classification'].should.equal('nested-configs')
    nested_config['Properties'].should.equal(args['Configurations'][1]['Properties'])

    attrs = cl['Ec2InstanceAttributes']
    attrs['AdditionalMasterSecurityGroups'].should.equal(
        args['Instances']['AdditionalMasterSecurityGroups'])
    attrs['AdditionalSlaveSecurityGroups'].should.equal(
        args['Instances']['AdditionalSlaveSecurityGroups'])
    attrs['Ec2AvailabilityZone'].should.equal('us-east-1a')
    attrs['Ec2KeyName'].should.equal(args['Instances']['Ec2KeyName'])
    attrs['Ec2SubnetId'].should.equal(args['Instances']['Ec2SubnetId'])
    attrs['EmrManagedMasterSecurityGroup'].should.equal(
        args['Instances']['EmrManagedMasterSecurityGroup'])
    attrs['EmrManagedSlaveSecurityGroup'].should.equal(
        args['Instances']['EmrManagedSlaveSecurityGroup'])
    attrs['IamInstanceProfile'].should.equal(args['JobFlowRole'])
    attrs['ServiceAccessSecurityGroup'].should.equal(
        args['Instances']['ServiceAccessSecurityGroup'])
    cl['Id'].should.equal(cluster_id)
    cl['LogUri'].should.equal(args['LogUri'])
    cl['MasterPublicDnsName'].should.be.a(six.string_types)
    cl['Name'].should.equal(args['Name'])
    cl['NormalizedInstanceHours'].should.equal(0)
    # cl['ReleaseLabel'].should.equal('emr-5.0.0')
    cl.shouldnt.have.key('RequestedAmiVersion')
    cl['RunningAmiVersion'].should.equal('1.0.0')
    # cl['SecurityConfiguration'].should.be.a(six.string_types)
    cl['ServiceRole'].should.equal(args['ServiceRole'])

    status = cl['Status']
    status['State'].should.equal('TERMINATED')
    # cluster['Status']['StateChangeReason']
    status['Timeline']['CreationDateTime'].should.be.a('datetime.datetime')
    # status['Timeline']['EndDateTime'].should.equal(datetime(2014, 1, 24, 2, 19, 46, tzinfo=pytz.utc))
    status['Timeline']['ReadyDateTime'].should.be.a('datetime.datetime')

    dict((t['Key'], t['Value']) for t in cl['Tags']).should.equal(
        dict((t['Key'], t['Value']) for t in args['Tags']))

    cl['TerminationProtected'].should.equal(False)
    cl['VisibleToAllUsers'].should.equal(True)


@mock_emr
def test_describe_cluster_not_found():
    conn = boto3.client('emr', region_name='us-east-1')
    raised = False
    try:
        cluster = conn.describe_cluster(ClusterId='DummyId')
    except ClientError as e:
        if e.response['Error']['Code'] == "ResourceNotFoundException":
            raised = True
    raised.should.equal(True)


@mock_emr
def test_describe_job_flows():
    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    expected = {}

    for idx in range(4):
        cluster_name = 'cluster' + str(idx)
        args['Name'] = cluster_name
        cluster_id = client.run_job_flow(**args)['JobFlowId']
        expected[cluster_id] = {
            'Id': cluster_id,
            'Name': cluster_name,
            'State': 'WAITING'
        }

    # need sleep since it appears the timestamp is always rounded to
    # the nearest second internally
    time.sleep(1)
    timestamp = datetime.now(pytz.utc)
    time.sleep(1)

    for idx in range(4, 6):
        cluster_name = 'cluster' + str(idx)
        args['Name'] = cluster_name
        cluster_id = client.run_job_flow(**args)['JobFlowId']
        client.terminate_job_flows(JobFlowIds=[cluster_id])
        expected[cluster_id] = {
            'Id': cluster_id,
            'Name': cluster_name,
            'State': 'TERMINATED'
        }

    resp = client.describe_job_flows()
    resp['JobFlows'].should.have.length_of(6)

    for cluster_id, y in expected.items():
        resp = client.describe_job_flows(JobFlowIds=[cluster_id])
        resp['JobFlows'].should.have.length_of(1)
        resp['JobFlows'][0]['JobFlowId'].should.equal(cluster_id)

    resp = client.describe_job_flows(JobFlowStates=['WAITING'])
    resp['JobFlows'].should.have.length_of(4)
    for x in resp['JobFlows']:
        x['ExecutionStatusDetail']['State'].should.equal('WAITING')

    resp = client.describe_job_flows(CreatedBefore=timestamp)
    resp['JobFlows'].should.have.length_of(4)

    resp = client.describe_job_flows(CreatedAfter=timestamp)
    resp['JobFlows'].should.have.length_of(2)


@mock_emr
def test_describe_job_flow():
    client = boto3.client('emr', region_name='us-east-1')

    args = deepcopy(run_job_flow_args)
    args['AmiVersion'] = '3.8.1'
    args['Instances'].update(
        {'Ec2KeyName': 'ec2keyname',
         'Ec2SubnetId': 'subnet-8be41cec',
         'HadoopVersion': '2.4.0'})
    args['VisibleToAllUsers'] = True

    cluster_id = client.run_job_flow(**args)['JobFlowId']

    jf = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]

    jf['AmiVersion'].should.equal(args['AmiVersion'])
    jf.shouldnt.have.key('BootstrapActions')
    esd = jf['ExecutionStatusDetail']
    esd['CreationDateTime'].should.be.a('datetime.datetime')
    # esd['EndDateTime'].should.be.a('datetime.datetime')
    # esd['LastStateChangeReason'].should.be.a(six.string_types)
    esd['ReadyDateTime'].should.be.a('datetime.datetime')
    esd['StartDateTime'].should.be.a('datetime.datetime')
    esd['State'].should.equal('WAITING')
    attrs = jf['Instances']
    attrs['Ec2KeyName'].should.equal(args['Instances']['Ec2KeyName'])
    attrs['Ec2SubnetId'].should.equal(args['Instances']['Ec2SubnetId'])
    attrs['HadoopVersion'].should.equal(args['Instances']['HadoopVersion'])
    attrs['InstanceCount'].should.equal(args['Instances']['InstanceCount'])
    for ig in attrs['InstanceGroups']:
        # ig['BidPrice']
        ig['CreationDateTime'].should.be.a('datetime.datetime')
        # ig['EndDateTime'].should.be.a('datetime.datetime')
        ig['InstanceGroupId'].should.be.a(six.string_types)
        ig['InstanceRequestCount'].should.be.a(int)
        ig['InstanceRole'].should.be.within(['MASTER', 'CORE'])
        ig['InstanceRunningCount'].should.be.a(int)
        ig['InstanceType'].should.be.within(['c3.medium', 'c3.xlarge'])
        # ig['LastStateChangeReason'].should.be.a(six.string_types)
        ig['Market'].should.equal('ON_DEMAND')
        ig['Name'].should.be.a(six.string_types)
        ig['ReadyDateTime'].should.be.a('datetime.datetime')
        ig['StartDateTime'].should.be.a('datetime.datetime')
        ig['State'].should.equal('RUNNING')
    attrs['KeepJobFlowAliveWhenNoSteps'].should.equal(True)
    # attrs['MasterInstanceId'].should.be.a(six.string_types)
    attrs['MasterInstanceType'].should.equal(
        args['Instances']['MasterInstanceType'])
    attrs['MasterPublicDnsName'].should.be.a(six.string_types)
    attrs['NormalizedInstanceHours'].should.equal(0)
    attrs['Placement']['AvailabilityZone'].should.equal(
        args['Instances']['Placement']['AvailabilityZone'])
    attrs['SlaveInstanceType'].should.equal(
        args['Instances']['SlaveInstanceType'])
    attrs['TerminationProtected'].should.equal(False)
    jf['JobFlowId'].should.equal(cluster_id)
    jf['JobFlowRole'].should.equal(args['JobFlowRole'])
    jf['LogUri'].should.equal(args['LogUri'])
    jf['Name'].should.equal(args['Name'])
    jf['ServiceRole'].should.equal(args['ServiceRole'])
    jf['Steps'].should.equal([])
    jf['SupportedProducts'].should.equal([])
    jf['VisibleToAllUsers'].should.equal(True)


@mock_emr
def test_list_clusters():
    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    expected = {}

    for idx in range(40):
        cluster_name = 'jobflow' + str(idx)
        args['Name'] = cluster_name
        cluster_id = client.run_job_flow(**args)['JobFlowId']
        expected[cluster_id] = {
            'Id': cluster_id,
            'Name': cluster_name,
            'NormalizedInstanceHours': 0,
            'State': 'WAITING'
        }

    # need sleep since it appears the timestamp is always rounded to
    # the nearest second internally
    time.sleep(1)
    timestamp = datetime.now(pytz.utc)
    time.sleep(1)

    for idx in range(40, 70):
        cluster_name = 'jobflow' + str(idx)
        args['Name'] = cluster_name
        cluster_id = client.run_job_flow(**args)['JobFlowId']
        client.terminate_job_flows(JobFlowIds=[cluster_id])
        expected[cluster_id] = {
            'Id': cluster_id,
            'Name': cluster_name,
            'NormalizedInstanceHours': 0,
            'State': 'TERMINATED'
        }

    args = {}
    while 1:
        resp = client.list_clusters(**args)
        clusters = resp['Clusters']
        len(clusters).should.be.lower_than_or_equal_to(50)
        for x in clusters:
            y = expected[x['Id']]
            x['Id'].should.equal(y['Id'])
            x['Name'].should.equal(y['Name'])
            x['NormalizedInstanceHours'].should.equal(
                y['NormalizedInstanceHours'])
            x['Status']['State'].should.equal(y['State'])
            x['Status']['Timeline'][
                'CreationDateTime'].should.be.a('datetime.datetime')
            if y['State'] == 'TERMINATED':
                x['Status']['Timeline'][
                    'EndDateTime'].should.be.a('datetime.datetime')
            else:
                x['Status']['Timeline'].shouldnt.have.key('EndDateTime')
            x['Status']['Timeline'][
                'ReadyDateTime'].should.be.a('datetime.datetime')
        marker = resp.get('Marker')
        if marker is None:
            break
        args = {'Marker': marker}

    resp = client.list_clusters(ClusterStates=['TERMINATED'])
    resp['Clusters'].should.have.length_of(30)
    for x in resp['Clusters']:
        x['Status']['State'].should.equal('TERMINATED')

    resp = client.list_clusters(CreatedBefore=timestamp)
    resp['Clusters'].should.have.length_of(40)

    resp = client.list_clusters(CreatedAfter=timestamp)
    resp['Clusters'].should.have.length_of(30)


@mock_emr
def test_run_job_flow():
    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    cluster_id = client.run_job_flow(**args)['JobFlowId']
    resp = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]
    resp['ExecutionStatusDetail']['State'].should.equal('WAITING')
    resp['JobFlowId'].should.equal(cluster_id)
    resp['Name'].should.equal(args['Name'])
    resp['Instances']['MasterInstanceType'].should.equal(
        args['Instances']['MasterInstanceType'])
    resp['Instances']['SlaveInstanceType'].should.equal(
        args['Instances']['SlaveInstanceType'])
    resp['LogUri'].should.equal(args['LogUri'])
    resp['VisibleToAllUsers'].should.equal(args['VisibleToAllUsers'])
    resp['Instances']['NormalizedInstanceHours'].should.equal(0)
    resp['Steps'].should.equal([])


@mock_emr
def test_run_job_flow_with_invalid_params():
    client = boto3.client('emr', region_name='us-east-1')
    with assert_raises(ClientError) as ex:
        # cannot set both AmiVersion and ReleaseLabel
        args = deepcopy(run_job_flow_args)
        args['AmiVersion'] = '2.4'
        args['ReleaseLabel'] = 'emr-5.0.0'
        client.run_job_flow(**args)
    ex.exception.response['Error']['Code'].should.equal('ValidationException')


@mock_emr
def test_run_job_flow_in_multiple_regions():
    regions = {}
    for region in ['us-east-1', 'eu-west-1']:
        client = boto3.client('emr', region_name=region)
        args = deepcopy(run_job_flow_args)
        args['Name'] = region
        cluster_id = client.run_job_flow(**args)['JobFlowId']
        regions[region] = {'client': client, 'cluster_id': cluster_id}

    for region in regions.keys():
        client = regions[region]['client']
        resp = client.describe_cluster(ClusterId=regions[region]['cluster_id'])
        resp['Cluster']['Name'].should.equal(region)


@mock_emr
def test_run_job_flow_with_new_params():
    client = boto3.client('emr', region_name='us-east-1')
    resp = client.run_job_flow(**run_job_flow_args)
    resp.should.have.key('JobFlowId')


@mock_emr
def test_run_job_flow_with_visible_to_all_users():
    client = boto3.client('emr', region_name='us-east-1')
    for expected in (True, False):
        args = deepcopy(run_job_flow_args)
        args['VisibleToAllUsers'] = expected
        resp = client.run_job_flow(**args)
        cluster_id = resp['JobFlowId']
        resp = client.describe_cluster(ClusterId=cluster_id)
        resp['Cluster']['VisibleToAllUsers'].should.equal(expected)


@mock_emr
def test_run_job_flow_with_instance_groups():
    input_groups = dict((g['Name'], g) for g in input_instance_groups)
    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    args['Instances'] = {'InstanceGroups': input_instance_groups}
    cluster_id = client.run_job_flow(**args)['JobFlowId']
    groups = client.list_instance_groups(ClusterId=cluster_id)[
        'InstanceGroups']
    for x in groups:
        y = input_groups[x['Name']]
        x.should.have.key('Id')
        x['RequestedInstanceCount'].should.equal(y['InstanceCount'])
        x['InstanceGroupType'].should.equal(y['InstanceRole'])
        x['InstanceType'].should.equal(y['InstanceType'])
        x['Market'].should.equal(y['Market'])
        if 'BidPrice' in y:
            x['BidPrice'].should.equal(y['BidPrice'])


@mock_emr
def test_run_job_flow_with_custom_ami():
    client = boto3.client('emr', region_name='us-east-1')

    with assert_raises(ClientError) as ex:
        # CustomAmiId available in Amazon EMR 5.7.0 and later
        args = deepcopy(run_job_flow_args)
        args['CustomAmiId'] = 'MyEmrCustomId'
        args['ReleaseLabel'] = 'emr-5.6.0'
        client.run_job_flow(**args)
    ex.exception.response['Error']['Code'].should.equal('ValidationException')
    ex.exception.response['Error']['Message'].should.equal('Custom AMI is not allowed')

    with assert_raises(ClientError) as ex:
        args = deepcopy(run_job_flow_args)
        args['CustomAmiId'] = 'MyEmrCustomId'
        args['AmiVersion'] = '3.8.1'
        client.run_job_flow(**args)
    ex.exception.response['Error']['Code'].should.equal('ValidationException')
    ex.exception.response['Error']['Message'].should.equal(
        'Custom AMI is not supported in this version of EMR')

    with assert_raises(ClientError) as ex:
        # AMI version and release label exception  raises before CustomAmi exception
        args = deepcopy(run_job_flow_args)
        args['CustomAmiId'] = 'MyEmrCustomId'
        args['ReleaseLabel'] = 'emr-5.6.0'
        args['AmiVersion'] = '3.8.1'
        client.run_job_flow(**args)
    ex.exception.response['Error']['Code'].should.equal('ValidationException')
    ex.exception.response['Error']['Message'].should.contain(
        'Only one AMI version and release label may be specified.')

    args = deepcopy(run_job_flow_args)
    args['CustomAmiId'] = 'MyEmrCustomAmi'
    args['ReleaseLabel'] = 'emr-5.7.0'
    cluster_id = client.run_job_flow(**args)['JobFlowId']
    resp = client.describe_cluster(ClusterId=cluster_id)
    resp['Cluster']['CustomAmiId'].should.equal('MyEmrCustomAmi')


@mock_emr
def test_set_termination_protection():
    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    args['Instances']['TerminationProtected'] = False
    resp = client.run_job_flow(**args)
    cluster_id = resp['JobFlowId']
    resp = client.describe_cluster(ClusterId=cluster_id)
    resp['Cluster']['TerminationProtected'].should.equal(False)

    for expected in (True, False):
        resp = client.set_termination_protection(JobFlowIds=[cluster_id],
                                                 TerminationProtected=expected)
        resp = client.describe_cluster(ClusterId=cluster_id)
        resp['Cluster']['TerminationProtected'].should.equal(expected)


@mock_emr
def test_set_visible_to_all_users():
    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    args['VisibleToAllUsers'] = False
    resp = client.run_job_flow(**args)
    cluster_id = resp['JobFlowId']
    resp = client.describe_cluster(ClusterId=cluster_id)
    resp['Cluster']['VisibleToAllUsers'].should.equal(False)

    for expected in (True, False):
        resp = client.set_visible_to_all_users(JobFlowIds=[cluster_id],
                                               VisibleToAllUsers=expected)
        resp = client.describe_cluster(ClusterId=cluster_id)
        resp['Cluster']['VisibleToAllUsers'].should.equal(expected)


@mock_emr
def test_terminate_job_flows():
    client = boto3.client('emr', region_name='us-east-1')

    resp = client.run_job_flow(**run_job_flow_args)
    cluster_id = resp['JobFlowId']
    resp = client.describe_cluster(ClusterId=cluster_id)
    resp['Cluster']['Status']['State'].should.equal('WAITING')

    resp = client.terminate_job_flows(JobFlowIds=[cluster_id])
    resp = client.describe_cluster(ClusterId=cluster_id)
    resp['Cluster']['Status']['State'].should.equal('TERMINATED')


# testing multiple end points for each feature

@mock_emr
def test_bootstrap_actions():
    bootstrap_actions = [
        {'Name': 'bs1',
         'ScriptBootstrapAction': {
             'Args': ['arg1', 'arg2'],
             'Path': 's3://path/to/script'}},
        {'Name': 'bs2',
         'ScriptBootstrapAction': {
             'Args': [],
             'Path': 's3://path/to/anotherscript'}}
    ]

    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    args['BootstrapActions'] = bootstrap_actions
    cluster_id = client.run_job_flow(**args)['JobFlowId']

    cl = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]
    for x, y in zip(cl['BootstrapActions'], bootstrap_actions):
        x['BootstrapActionConfig'].should.equal(y)

    resp = client.list_bootstrap_actions(ClusterId=cluster_id)
    for x, y in zip(resp['BootstrapActions'], bootstrap_actions):
        x['Name'].should.equal(y['Name'])
        if 'Args' in y['ScriptBootstrapAction']:
            x['Args'].should.equal(y['ScriptBootstrapAction']['Args'])
        x['ScriptPath'].should.equal(y['ScriptBootstrapAction']['Path'])


@mock_emr
def test_instance_groups():
    input_groups = dict((g['Name'], g) for g in input_instance_groups)

    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    for key in ['MasterInstanceType', 'SlaveInstanceType', 'InstanceCount']:
        del args['Instances'][key]
    args['Instances']['InstanceGroups'] = input_instance_groups[:2]
    cluster_id = client.run_job_flow(**args)['JobFlowId']

    jf = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]
    base_instance_count = jf['Instances']['InstanceCount']

    client.add_instance_groups(
        JobFlowId=cluster_id, InstanceGroups=input_instance_groups[2:])

    jf = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]
    jf['Instances']['InstanceCount'].should.equal(
        sum(g['InstanceCount'] for g in input_instance_groups))
    for x in jf['Instances']['InstanceGroups']:
        y = input_groups[x['Name']]
        if hasattr(y, 'BidPrice'):
            x['BidPrice'].should.equal('BidPrice')
        x['CreationDateTime'].should.be.a('datetime.datetime')
        # x['EndDateTime'].should.be.a('datetime.datetime')
        x.should.have.key('InstanceGroupId')
        x['InstanceRequestCount'].should.equal(y['InstanceCount'])
        x['InstanceRole'].should.equal(y['InstanceRole'])
        x['InstanceRunningCount'].should.equal(y['InstanceCount'])
        x['InstanceType'].should.equal(y['InstanceType'])
        # x['LastStateChangeReason'].should.equal(y['LastStateChangeReason'])
        x['Market'].should.equal(y['Market'])
        x['Name'].should.equal(y['Name'])
        x['ReadyDateTime'].should.be.a('datetime.datetime')
        x['StartDateTime'].should.be.a('datetime.datetime')
        x['State'].should.equal('RUNNING')

    groups = client.list_instance_groups(ClusterId=cluster_id)[
        'InstanceGroups']
    for x in groups:
        y = input_groups[x['Name']]
        if hasattr(y, 'BidPrice'):
            x['BidPrice'].should.equal('BidPrice')
        # Configurations
        # EbsBlockDevices
        # EbsOptimized
        x.should.have.key('Id')
        x['InstanceGroupType'].should.equal(y['InstanceRole'])
        x['InstanceType'].should.equal(y['InstanceType'])
        x['Market'].should.equal(y['Market'])
        x['Name'].should.equal(y['Name'])
        x['RequestedInstanceCount'].should.equal(y['InstanceCount'])
        x['RunningInstanceCount'].should.equal(y['InstanceCount'])
        # ShrinkPolicy
        x['Status']['State'].should.equal('RUNNING')
        x['Status']['StateChangeReason']['Code'].should.be.a(six.string_types)
        # x['Status']['StateChangeReason']['Message'].should.be.a(six.string_types)
        x['Status']['Timeline'][
            'CreationDateTime'].should.be.a('datetime.datetime')
        # x['Status']['Timeline']['EndDateTime'].should.be.a('datetime.datetime')
        x['Status']['Timeline'][
            'ReadyDateTime'].should.be.a('datetime.datetime')

    igs = dict((g['Name'], g) for g in groups)
    client.modify_instance_groups(
        InstanceGroups=[
            {'InstanceGroupId': igs['task-1']['Id'],
             'InstanceCount': 2},
            {'InstanceGroupId': igs['task-2']['Id'],
             'InstanceCount': 3}])
    jf = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]
    jf['Instances']['InstanceCount'].should.equal(base_instance_count + 5)
    igs = dict((g['Name'], g) for g in jf['Instances']['InstanceGroups'])
    igs['task-1']['InstanceRunningCount'].should.equal(2)
    igs['task-2']['InstanceRunningCount'].should.equal(3)


@mock_emr
def test_steps():
    input_steps = [{
        'HadoopJarStep': {
            'Args': [
                'hadoop-streaming',
                '-files', 's3://elasticmapreduce/samples/wordcount/wordSplitter.py#wordSplitter.py',
                '-mapper', 'python wordSplitter.py',
                '-input', 's3://elasticmapreduce/samples/wordcount/input',
                '-output', 's3://output_bucket/output/wordcount_output',
                '-reducer', 'aggregate'
            ],
            'Jar': 'command-runner.jar',
        },
        'Name': 'My wordcount example',
    }, {
        'HadoopJarStep': {
            'Args': [
                'hadoop-streaming',
                '-files', 's3://elasticmapreduce/samples/wordcount/wordSplitter2.py#wordSplitter2.py',
                '-mapper', 'python wordSplitter2.py',
                '-input', 's3://elasticmapreduce/samples/wordcount/input2',
                '-output', 's3://output_bucket/output/wordcount_output2',
                '-reducer', 'aggregate'
            ],
            'Jar': 'command-runner.jar',
        },
        'Name': 'My wordcount example2',
    }]

    # TODO: implementation and test for cancel_steps

    client = boto3.client('emr', region_name='us-east-1')
    args = deepcopy(run_job_flow_args)
    args['Steps'] = [input_steps[0]]
    cluster_id = client.run_job_flow(**args)['JobFlowId']

    jf = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]
    jf['Steps'].should.have.length_of(1)

    client.add_job_flow_steps(JobFlowId=cluster_id, Steps=[input_steps[1]])

    jf = client.describe_job_flows(JobFlowIds=[cluster_id])['JobFlows'][0]
    jf['Steps'].should.have.length_of(2)
    for idx, (x, y) in enumerate(zip(jf['Steps'], input_steps)):
        x['ExecutionStatusDetail'].should.have.key('CreationDateTime')
        # x['ExecutionStatusDetail'].should.have.key('EndDateTime')
        # x['ExecutionStatusDetail'].should.have.key('LastStateChangeReason')
        # x['ExecutionStatusDetail'].should.have.key('StartDateTime')
        x['ExecutionStatusDetail']['State'].should.equal(
            'STARTING' if idx == 0 else 'PENDING')
        x['StepConfig']['ActionOnFailure'].should.equal('TERMINATE_CLUSTER')
        x['StepConfig']['HadoopJarStep'][
            'Args'].should.equal(y['HadoopJarStep']['Args'])
        x['StepConfig']['HadoopJarStep'][
            'Jar'].should.equal(y['HadoopJarStep']['Jar'])
        if 'MainClass' in y['HadoopJarStep']:
            x['StepConfig']['HadoopJarStep']['MainClass'].should.equal(
                y['HadoopJarStep']['MainClass'])
        if 'Properties' in y['HadoopJarStep']:
            x['StepConfig']['HadoopJarStep']['Properties'].should.equal(
                y['HadoopJarStep']['Properties'])
        x['StepConfig']['Name'].should.equal(y['Name'])

    expected = dict((s['Name'], s) for s in input_steps)

    steps = client.list_steps(ClusterId=cluster_id)['Steps']
    steps.should.have.length_of(2)
    for x in steps:
        y = expected[x['Name']]
        x['ActionOnFailure'].should.equal('TERMINATE_CLUSTER')
        x['Config']['Args'].should.equal(y['HadoopJarStep']['Args'])
        x['Config']['Jar'].should.equal(y['HadoopJarStep']['Jar'])
        # x['Config']['MainClass'].should.equal(y['HadoopJarStep']['MainClass'])
        # Properties
        x['Id'].should.be.a(six.string_types)
        x['Name'].should.equal(y['Name'])
        x['Status']['State'].should.be.within(['STARTING', 'PENDING'])
        # StateChangeReason
        x['Status']['Timeline'][
            'CreationDateTime'].should.be.a('datetime.datetime')
        # x['Status']['Timeline']['EndDateTime'].should.be.a('datetime.datetime')
        # x['Status']['Timeline']['StartDateTime'].should.be.a('datetime.datetime')

        x = client.describe_step(ClusterId=cluster_id, StepId=x['Id'])['Step']
        x['ActionOnFailure'].should.equal('TERMINATE_CLUSTER')
        x['Config']['Args'].should.equal(y['HadoopJarStep']['Args'])
        x['Config']['Jar'].should.equal(y['HadoopJarStep']['Jar'])
        # x['Config']['MainClass'].should.equal(y['HadoopJarStep']['MainClass'])
        # Properties
        x['Id'].should.be.a(six.string_types)
        x['Name'].should.equal(y['Name'])
        x['Status']['State'].should.be.within(['STARTING', 'PENDING'])
        # StateChangeReason
        x['Status']['Timeline'][
            'CreationDateTime'].should.be.a('datetime.datetime')
        # x['Status']['Timeline']['EndDateTime'].should.be.a('datetime.datetime')
        # x['Status']['Timeline']['StartDateTime'].should.be.a('datetime.datetime')

    step_id = steps[0]['Id']
    steps = client.list_steps(ClusterId=cluster_id, StepIds=[step_id])['Steps']
    steps.should.have.length_of(1)
    steps[0]['Id'].should.equal(step_id)

    steps = client.list_steps(ClusterId=cluster_id,
                              StepStates=['STARTING'])['Steps']
    steps.should.have.length_of(1)
    steps[0]['Id'].should.equal(step_id)


@mock_emr
def test_tags():
    input_tags = [{'Key': 'newkey1', 'Value': 'newval1'},
                  {'Key': 'newkey2', 'Value': 'newval2'}]

    client = boto3.client('emr', region_name='us-east-1')
    cluster_id = client.run_job_flow(**run_job_flow_args)['JobFlowId']

    client.add_tags(ResourceId=cluster_id, Tags=input_tags)
    resp = client.describe_cluster(ClusterId=cluster_id)['Cluster']
    resp['Tags'].should.have.length_of(2)
    dict((t['Key'], t['Value']) for t in resp['Tags']).should.equal(
        dict((t['Key'], t['Value']) for t in input_tags))

    client.remove_tags(ResourceId=cluster_id, TagKeys=[
                       t['Key'] for t in input_tags])
    resp = client.describe_cluster(ClusterId=cluster_id)['Cluster']
    resp['Tags'].should.equal([])



import base64
import os
import random
import string

import six

ACCOUNT_SPECIFIC_ACCESS_KEY_PREFIX = "8NWMTLYQ"
SESSION_TOKEN_PREFIX = "FQoGZXIvYXdzEBYaD"


def random_access_key_id():
    return ACCOUNT_SPECIFIC_ACCESS_KEY_PREFIX + ''.join(six.text_type(
        random.choice(
            string.ascii_uppercase + string.digits
        )) for _ in range(8)
    )


def random_secret_access_key():
    return base64.b64encode(os.urandom(30)).decode()


def random_session_token():
    return SESSION_TOKEN_PREFIX + base64.b64encode(os.urandom(266))[len(SESSION_TOKEN_PREFIX):].decode()
